@@ -4,7 +4,13 @@
 #ifndef CEPH_CRUSH_WRAPPER_H
 #define CEPH_CRUSH_WRAPPER_H
 
-#define BUG_ON(x) assert(!(x))
+#include <stdlib.h>
+#include <map>
+#include <set>
+#include <string>
+
+#include <iostream> //for testing, remove
+
 #include "include/types.h"
 
 extern "C" {
@@ -17,12 +23,11 @@ extern "C" {
 #include "include/err.h"
 #include "include/encoding.h"
 
-#include <stdlib.h>
-#include <map>
-#include <set>
-#include <string>
 
-#include <iostream> //for testing, remove
+#include "common/Mutex.h"
+
+#include "include/assert.h"
+#define BUG_ON(x) assert(!(x))
 
 namespace ceph {
   class Formatter;
@@ -45,15 +50,16 @@ inline static void decode(crush_rule_step &s, bufferlist::iterator &p)
 
 using namespace std;
 class CrushWrapper {
+  mutable Mutex mapper_lock;
 public:
   struct crush_map *crush;
-  std::map<int, string> type_map; /* bucket/device type names */
-  std::map<int, string> name_map; /* bucket/device names */
-  std::map<int, string> rule_name_map;
+  std::map<int32_t, string> type_map; /* bucket/device type names */
+  std::map<int32_t, string> name_map; /* bucket/device names */
+  std::map<int32_t, string> rule_name_map;
 
   /* reverse maps */
-  mutable bool have_rmaps;
-  mutable std::map<string, int> type_rmap, name_rmap, rule_name_rmap;
+  bool have_rmaps;
+  std::map<string, int> type_rmap, name_rmap, rule_name_rmap;
 
 private:
   void build_rmaps() {
@@ -73,7 +79,8 @@ public:
   CrushWrapper(const CrushWrapper& other);
   const CrushWrapper& operator=(const CrushWrapper& other);
 
-  CrushWrapper() : crush(0), have_rmaps(false) {
+  CrushWrapper() : mapper_lock("CrushWrapper::mapper_lock"),
+		   crush(0), have_rmaps(false) {
     create();
   }
   ~CrushWrapper() {
@@ -86,6 +93,29 @@ public:
     if (crush)
       crush_destroy(crush);
     crush = crush_create();
+    assert(crush);
+  }
+
+  // tunables
+  int get_choose_local_tries() const {
+    return crush->choose_local_tries;
+  }
+  void set_choose_local_tries(int n) {
+    crush->choose_local_tries = n;
+  }
+
+  int get_choose_local_fallback_tries() const {
+    return crush->choose_local_fallback_tries;
+  }
+  void set_choose_local_fallback_tries(int n) {
+    crush->choose_local_fallback_tries = n;
+  }
+
+  int get_choose_total_tries() const {
+    return crush->choose_total_tries;
+  }
+  void set_choose_total_tries(int n) {
+    crush->choose_total_tries = n;
   }
 
   // bucket types
@@ -191,6 +221,30 @@ public:
     return ret;
   }
 
+
+  /**
+   * returns the (type, name) of the parent bucket of id
+   */
+  pair<string,string> get_immediate_parent(int id);
+
+  /**
+   * get the fully qualified location of a device by successively finding
+   * parents beginning at ID and ending at highest type number specified in
+   * the CRUSH map which assumes that if device foo is under device bar, the
+   * type_id of foo < bar where type_id is the integer specified in the CRUSH map
+   *
+   * returns the location in the form of (type=foo) where type is a type of bucket
+   * specified in the CRUSH map and foo is a name specified in the CRUSH map
+   */
+  map<string, string> get_full_location(int id);
+
+  /**
+   * returns (type_id, type) of all parent buckets between id and
+   * default, can be used to check for anomolous CRUSH maps
+   */
+  map<int, string> get_parent_hierarchy(int id);
+
+
   /**
    * insert an item into the map at a specific position
    *
@@ -223,6 +277,19 @@ public:
    */
   int insert_item(CephContext *cct, int id, float weight, string name, map<string,string>& loc);
 
+  /**
+   * move a bucket in the hierarchy to the given location
+   *
+   * This has the same location and ancestor creation behavior as
+   * insert_item(), but will relocate the specified existing bucket.
+   *
+   * @param cct cct
+   * @param id bucket id
+   * @param loc location (map of type to bucket names)
+   * @return 0 for success, negative on error
+   */
+  int move_bucket(CephContext *cct, int id, map<string,string>& loc);
+
   /**
    * add or update an item's position in the map
    *
@@ -252,6 +319,10 @@ public:
   }
   void reweight(CephContext *cct);
 
+  /// check if item id is present in the map hierarchy
+  bool check_item_present(int id);
+
+
   /*** devices ***/
   int get_max_devices() const {
     if (!crush) return 0;
@@ -332,6 +403,7 @@ public:
   int add_rule(int len, int ruleset, int type, int minsize, int maxsize, int ruleno) {
     if (!crush) return -ENOENT;
     crush_rule *n = crush_make_rule(len, ruleset, type, minsize, maxsize);
+    assert(n);
     ruleno = crush_add_rule(crush, n, ruleno);
     return ruleno;
   }
@@ -389,6 +461,51 @@ private:
       return (crush_bucket *)(-ENOENT);
     return ret;
   }
+  /**
+   * detach a bucket from its parent and adjust the parent weight
+   *
+   * returns the weight of the detached bucket
+   **/
+  int detach_bucket(CephContext *cct, int item){
+    if (!crush)
+      return (-EINVAL);
+
+    if (item > 0)
+      return (-EINVAL);
+
+    // check that the bucket that we want to detach exists
+    assert( get_bucket(item) );
+
+    // get the bucket's weight
+    crush_bucket *b = get_bucket(item);
+    unsigned bucket_weight = b->weight;
+
+    // zero out the bucket weight
+    adjust_item_weight(cct, item, 0);
+
+    // get where the bucket is located
+    pair<string, string> bucket_location = get_immediate_parent(item);
+
+    // get the id of the parent bucket
+    int parent_id = get_item_id( (bucket_location.second).c_str() );
+
+    // get the parent bucket
+    crush_bucket *parent_bucket = get_bucket(parent_id);
+
+    // remove the bucket from the parent
+    crush_bucket_remove_item(parent_bucket, item);
+
+    // check that we're happy
+    int test_weight = 0;
+    map<string,string> test_location;
+    test_location[ bucket_location.first ] = (bucket_location.second);
+
+    bool successful_detach = !(check_item_loc(cct, item, test_location, &test_weight));
+    assert(successful_detach);
+    assert(test_weight == 0);
+
+    return bucket_weight;
+  }
 
 public:
   int get_max_buckets() const {
@@ -452,6 +569,7 @@ public:
   int add_bucket(int bucketno, int alg, int hash, int type, int size,
 		 int *items, int *weights) {
     crush_bucket *b = crush_make_bucket(alg, hash, type, size, items, weights);
+    assert(b);
     return crush_add_bucket(crush, bucketno, b);
   }
   
@@ -460,6 +578,26 @@ public:
     crush_finalize(crush);
   }
 
+  void start_choose_profile() {
+    free(crush->choose_tries);
+    crush->choose_tries = (__u32 *)malloc(sizeof(*crush->choose_tries) * crush->choose_total_tries);
+    memset(crush->choose_tries, 0,
+	   sizeof(*crush->choose_tries) * crush->choose_total_tries);
+  }
+  void stop_choose_profile() {
+    free(crush->choose_tries);
+    crush->choose_tries = 0;
+  }
+
+  int get_choose_profile(__u32 **vec) {
+    if (crush->choose_tries) {
+      *vec = crush->choose_tries;
+      return crush->choose_total_tries;
+    }
+    return 0;
+  }
+
+
   void set_max_devices(int m) {
     crush->max_devices = m;
   }
@@ -470,6 +608,7 @@ public:
   }
   void do_rule(int rule, int x, vector<int>& out, int maxout,
 	       const vector<__u32>& weight) const {
+    Mutex::Locker l(mapper_lock);
     int rawout[maxout];
     int numrep = crush_do_rule(crush, rule, x, rawout, maxout, &weight[0], weight.size());
     if (numrep < 0)
