@@ -15,14 +15,15 @@
 #ifndef CEPH_ENCODING_H
 #define CEPH_ENCODING_H
 
+#include <tr1/memory>
+
 #include "inttypes.h"
 #include "byteorder.h"
 #include "buffer.h"
+#include "assert.h"
 
 using namespace ceph;
 
-#include <tr1/memory>
-
 /*
  * Notes on feature encoding:
  *
@@ -58,7 +59,7 @@ inline void decode_raw(T& t, bufferlist::iterator &p)
 
 #define WRITE_RAW_ENCODER(type)						\
   inline void encode(const type &v, bufferlist& bl, uint64_t features=0) { encode_raw(v, bl); } \
-  inline void decode(type &v, bufferlist::iterator& p) { decode_raw(v, p); }
+  inline void decode(type &v, bufferlist::iterator& p) { __ASSERT_FUNCTION decode_raw(v, p); }
 
 WRITE_RAW_ENCODER(__u8)
 WRITE_RAW_ENCODER(__s8)
@@ -438,6 +439,14 @@ inline void decode(std::vector<T*>& v, bufferlist::iterator& p)
 */
 // vector
 template<class T>
+inline void encode(const std::vector<T>& v, bufferlist& bl, uint64_t features)
+{
+  __u32 n = v.size();
+  encode(n, bl);
+  for (typename std::vector<T>::const_iterator p = v.begin(); p != v.end(); ++p)
+    encode(*p, bl, features);
+}
+template<class T>
 inline void encode(const std::vector<T>& v, bufferlist& bl)
 {
   __u32 n = v.size();
@@ -794,6 +803,9 @@ inline void decode(std::deque<T>& ls, bufferlist::iterator& p)
 #define DECODE_START_LEGACY_COMPAT_LEN_32(v, compatv, lenv, bl)		\
   __DECODE_START_LEGACY_COMPAT_LEN(v, compatv, lenv, 3, bl)
 
+#define DECODE_START_LEGACY_COMPAT_LEN_16(v, compatv, lenv, bl)		\
+  __DECODE_START_LEGACY_COMPAT_LEN(v, compatv, lenv, 1, bl)
+
 /**
  * finish decode block
  *
