@@ -12,6 +12,11 @@
  * 
  */
 
+
+#include <sstream>
+#include <stdlib.h>
+#include <signal.h>
+
 #include "Monitor.h"
 
 #include "osd/OSDMap.h"
@@ -28,8 +33,6 @@
 #include "messages/MGenericMessage.h"
 #include "messages/MMonCommand.h"
 #include "messages/MMonCommandAck.h"
-#include "messages/MMonObserve.h"
-#include "messages/MMonObserveNotify.h"
 #include "messages/MMonProbe.h"
 #include "messages/MMonJoin.h"
 #include "messages/MMonPaxos.h"
@@ -51,6 +54,7 @@
 
 #include "include/color.h"
 #include "include/ceph_fs.h"
+#include "include/str_list.h"
 
 #include "OSDMonitor.h"
 #include "MDSMonitor.h"
@@ -62,12 +66,10 @@
 #include "osd/OSDMap.h"
 
 #include "auth/AuthSupported.h"
+#include "auth/KeyRing.h"
 
 #include "common/config.h"
-
-#include <sstream>
-#include <stdlib.h>
-#include <signal.h>
+#include "include/assert.h"
 
 #define dout_subsys ceph_subsys_mon
 #undef dout_prefix
@@ -94,10 +96,11 @@ Monitor::Monitor(CephContext* cct_, string nm, MonitorStore *s, Messenger *m, Mo
   messenger(m),
   lock("Monitor::lock"),
   timer(cct_, lock),
+  has_ever_joined(false),
   logger(NULL), cluster_logger(NULL), cluster_logger_registered(false),
   monmap(map),
-  clog(cct_, messenger, monmap, NULL, LogClient::FLAG_MON),
-  key_server(cct),
+  clog(cct_, messenger, monmap, LogClient::FLAG_MON),
+  key_server(cct, &keyring),
   auth_supported(cct),
   store(s),
   
@@ -105,6 +108,7 @@ Monitor::Monitor(CephContext* cct_, string nm, MonitorStore *s, Messenger *m, Mo
   
   elector(this),
   leader(0),
+  quorum_features(0),
   probe_timeout_event(NULL),
 
   paxos(PAXOS_NUM), paxos_service(PAXOS_NUM),
@@ -196,21 +200,23 @@ class AdminHook : public AdminSocketHook {
   Monitor *mon;
 public:
   AdminHook(Monitor *m) : mon(m) {}
-  bool call(std::string command, bufferlist& out) {
+  bool call(std::string command, std::string args, bufferlist& out) {
     stringstream ss;
-    mon->do_admin_command(command, ss);
+    mon->do_admin_command(command, args, ss);
     out.append(ss);
     return true;
   }
 };
 
-void Monitor::do_admin_command(string command, ostream& ss)
+void Monitor::do_admin_command(string command, string args, ostream& ss)
 {
   Mutex::Locker l(lock);
   if (command == "mon_status")
     _mon_status(ss);
   else if (command == "quorum_status")
     _quorum_status(ss);
+  else if (command.find("add_bootstrap_peer_hint") == 0)
+    _add_bootstrap_peer_hint(command, args, ss);
   else
     assert(0 == "bad AdminSocket command binding");
 }
@@ -222,12 +228,10 @@ void Monitor::handle_signal(int signum)
   shutdown();
 }
 
-void Monitor::init()
+int Monitor::init()
 {
   lock.Lock();
 
-  rank = monmap->get_rank(messenger->get_myaddr());
-  
   dout(1) << "init fsid " << monmap->fsid << dendl;
   
   assert(!logger);
@@ -279,6 +283,25 @@ void Monitor::init()
     dout(10) << "features " << features << dendl;
   }
 
+  // have we ever joined a quorum?
+  has_ever_joined = store->exists_bl_ss("joined");
+  dout(10) << "has_ever_joined = " << (int)has_ever_joined << dendl;
+
+  if (!has_ever_joined) {
+    // impose initial quorum restrictions?
+    list<string> initial_members;
+    get_str_list(g_conf->mon_initial_members, initial_members);
+
+    if (initial_members.size()) {
+      dout(1) << " initial_members " << initial_members << ", filtering seed monmap" << dendl;
+
+      monmap->set_initial_members(g_ceph_context, initial_members, name, messenger->get_myaddr(),
+				  &extra_probe_peers);
+
+      dout(10) << " monmap is " << *monmap << dendl;
+    }
+  }
+
   // init paxos
   for (int i = 0; i < PAXOS_NUM; ++i) {
     paxos[i]->init();
@@ -296,21 +319,46 @@ void Monitor::init()
     KeyRing keyring;
     bufferlist::iterator p = bl.begin();
     ::decode(keyring, p);
-    key_server.bootstrap_keyring(keyring);
+    extract_save_mon_key(keyring);
+  }
+
+  ostringstream os;
+  os << g_conf->mon_data << "/keyring";
+  int r = keyring.load(cct, os.str());
+  if (r < 0) {
+    EntityName mon_name;
+    mon_name.set_type(CEPH_ENTITY_TYPE_MON);
+    EntityAuth mon_key;
+    if (key_server.get_auth(mon_name, mon_key)) {
+      dout(1) << "copying mon. key from old db to external keyring" << dendl;
+      keyring.add(mon_name, mon_key);
+      bufferlist bl;
+      keyring.encode_plaintext(bl);
+      store->put_bl_ss(bl, "keyring", NULL);
+    } else {
+      derr << "unable to load initial keyring " << g_conf->keyring << dendl;
+      return r;
+    }
   }
 
   admin_hook = new AdminHook(this);
   AdminSocket* admin_socket = cct->get_admin_socket();
-  int r = admin_socket->register_command("mon_status", admin_hook,
-					 "show current monitor status");
+
+  // unlock while registering to avoid mon_lock -> admin socket lock dependency.
+  lock.Unlock();
+  r = admin_socket->register_command("mon_status", admin_hook,
+				     "show current monitor status");
   assert(r == 0);
   r = admin_socket->register_command("quorum_status", admin_hook,
 					 "show current quorum status");
   assert(r == 0);
+  r = admin_socket->register_command("add_bootstrap_peer_hint", admin_hook,
+				     "add peer address as potential bootstrap peer for cluster bringup");
+  assert(r == 0);
+  lock.Lock();
 
   // i'm ready!
   messenger->add_dispatcher_tail(this);
-  messenger->add_dispatcher_head(&clog);
   
   // start ticker
   timer.init();
@@ -319,6 +367,7 @@ void Monitor::init()
   bootstrap();
   
   lock.Unlock();
+  return 0;
 }
 
 void Monitor::register_cluster_logger()
@@ -401,7 +450,7 @@ void Monitor::bootstrap()
   int newrank = monmap->get_rank(messenger->get_myaddr());
   if (newrank < 0 && rank >= 0) {
     // was i ever part of the quorum?
-    if (store->exists_bl_ss("joined")) {
+    if (has_ever_joined) {
       dout(0) << " removed from monmap, suicide." << dendl;
       exit(0);
     }
@@ -429,14 +478,49 @@ void Monitor::bootstrap()
   reset_probe_timeout();
 
   // i'm outside the quorum
-  outside_quorum.insert(name);
+  if (monmap->contains(name))
+    outside_quorum.insert(name);
 
   // probe monitors
   dout(10) << "probing other monitors" << dendl;
   for (unsigned i = 0; i < monmap->size(); i++) {
     if ((int)i != rank)
-      messenger->send_message(new MMonProbe(monmap->fsid, MMonProbe::OP_PROBE, name), monmap->get_inst(i));
+      messenger->send_message(new MMonProbe(monmap->fsid, MMonProbe::OP_PROBE, name, has_ever_joined),
+			      monmap->get_inst(i));
   }
+  for (set<entity_addr_t>::iterator p = extra_probe_peers.begin();
+       p != extra_probe_peers.end();
+       ++p) {
+    if (*p != messenger->get_myaddr()) {
+      entity_inst_t i;
+      i.name = entity_name_t::MON(-1);
+      i.addr = *p;
+      messenger->send_message(new MMonProbe(monmap->fsid, MMonProbe::OP_PROBE, name, has_ever_joined), i);
+    }
+  }
+}
+
+void Monitor::_add_bootstrap_peer_hint(string cmd, string args, ostream& ss)
+{
+  dout(10) << "_add_bootstrap_peer_hint '" << cmd << "' '" << args << "'" << dendl;
+
+  entity_addr_t addr;
+  const char *end = 0;
+  if (!addr.parse(args.c_str(), &end)) {
+    ss << "failed to parse addr '" << args << "'; syntax is 'add_bootstrap_peer_hint ip[:port]'";
+    return;
+  }
+
+  if (is_leader() || is_peon()) {
+    ss << "mon already active; ignoring bootstrap hint";
+    return;
+  }
+
+  if (addr.get_port() == 0)
+    addr.set_port(CEPH_MON_PORT);
+
+  extra_probe_peers.insert(addr);
+  ss << "adding peer " << addr << " to list: " << extra_probe_peers;
 }
 
 // called by bootstrap(), or on leader|peon -> electing
@@ -524,36 +608,76 @@ void Monitor::handle_probe(MMonProbe *m)
 void Monitor::handle_probe_probe(MMonProbe *m)
 {
   dout(10) << "handle_probe_probe " << m->get_source_inst() << *m << dendl;
-  MMonProbe *r = new MMonProbe(monmap->fsid, MMonProbe::OP_REPLY, name);
+  MMonProbe *r = new MMonProbe(monmap->fsid, MMonProbe::OP_REPLY, name, has_ever_joined);
   r->name = name;
   r->quorum = quorum;
-  monmap->encode(r->monmap_bl);
+  monmap->encode(r->monmap_bl, m->get_connection()->get_features());
   for (vector<Paxos*>::iterator p = paxos.begin(); p != paxos.end(); ++p)
     r->paxos_versions[(*p)->get_machine_name()] = (*p)->get_version();
   messenger->send_message(r, m->get_connection());
+
+  // did we discover a peer here?
+  if (!monmap->contains(m->get_source_addr())) {
+    dout(1) << " adding peer " << m->get_source_addr() << " to list of hints" << dendl;
+    extra_probe_peers.insert(m->get_source_addr());
+  }
+
   m->put();
 }
 
 void Monitor::handle_probe_reply(MMonProbe *m)
 {
   dout(10) << "handle_probe_reply " << m->get_source_inst() << *m << dendl;
+  dout(10) << " monmap is " << *monmap << dendl;
 
   if (!is_probing()) {
     m->put();
     return;
   }
 
-  // newer map?
-  MonMap *newmap = new MonMap;
-  newmap->decode(m->monmap_bl);
-  if (newmap->get_epoch() > monmap->get_epoch()) {
-    dout(10) << " got new monmap epoch " << newmap->get_epoch()
-	     << " > my " << monmap->get_epoch() << dendl;
-    monmap->decode(m->monmap_bl);
-    m->put();
+  // newer map, or they've joined a quorum and we haven't?
+  bufferlist mybl;
+  monmap->encode(mybl, m->get_connection()->get_features());
+  // make sure it's actually different; the checks below err toward
+  // taking the other guy's map, which could cause us to loop.
+  if (!mybl.contents_equal(m->monmap_bl)) {
+    MonMap *newmap = new MonMap;
+    newmap->decode(m->monmap_bl);
+    if (m->has_ever_joined && (newmap->get_epoch() > monmap->get_epoch() ||
+			       !has_ever_joined)) {
+      dout(10) << " got newer/committed monmap epoch " << newmap->get_epoch()
+	       << ", mine was " << monmap->get_epoch() << dendl;
+      delete newmap;
+      monmap->decode(m->monmap_bl);
+      m->put();
 
-    bootstrap();
-    return;
+      bootstrap();
+      return;
+    }
+    delete newmap;
+  }
+
+  // rename peer?
+  string peer_name = monmap->get_name(m->get_source_addr());
+  if (monmap->get_epoch() == 0 && peer_name.find("noname-") == 0) {
+    dout(10) << " renaming peer " << m->get_source_addr() << " "
+	     << peer_name << " -> " << m->name << " in my monmap"
+	     << dendl;
+    monmap->rename(peer_name, m->name);
+  } else {
+    dout(10) << " peer name is " << peer_name << dendl;
+  }
+
+  // new initial peer?
+  if (monmap->contains(m->name)) {
+    if (monmap->get_addr(m->name).is_blank_ip()) {
+      dout(1) << " learned initial mon " << m->name << " addr " << m->get_source_addr() << dendl;
+      monmap->set_addr(m->name, m->get_source_addr());
+      m->put();
+
+      bootstrap();
+      return;
+    }
   }
 
   // is there an existing quorum?
@@ -589,11 +713,12 @@ void Monitor::handle_probe_reply(MMonProbe *m)
       }
     }
     if (ok) {
-      if (monmap->contains(name)) {
+      if (monmap->contains(name) &&
+	  !monmap->get_addr(name).is_blank_ip()) {
 	// i'm part of the cluster; just initiate a new election
 	start_election();
       } else {
-	dout(10) << " ready to join, but i'm not in the monmap, trying to join" << dendl;
+	dout(10) << " ready to join, but i'm not in the monmap or my addr is blank, trying to join" << dendl;
 	messenger->send_message(new MMonJoin(monmap->fsid, name, messenger->get_myaddr()),
 				monmap->get_inst(*m->quorum.begin()));
       }
@@ -604,14 +729,21 @@ void Monitor::handle_probe_reply(MMonProbe *m)
     }
   } else {
     // not part of a quorum
-    outside_quorum.insert(m->name);
+    if (monmap->contains(m->name))
+      outside_quorum.insert(m->name);
+    else
+      dout(10) << " mostly ignoring mon." << m->name << ", not part of monmap" << dendl;
 
     unsigned need = monmap->size() / 2 + 1;
     dout(10) << " outside_quorum now " << outside_quorum << ", need " << need << dendl;
 
     if (outside_quorum.size() >= need) {
-      dout(10) << " that's enough to form a new quorum, calling election" << dendl;
-      start_election();
+      if (outside_quorum.count(name)) {
+	dout(10) << " that's enough to form a new quorum, calling election" << dendl;
+	start_election();
+      } else {
+	dout(10) << " that's enough to form a new quorum, but it does not include me; waiting" << dendl;
+      }
     } else {
       dout(10) << " that's not yet enough for a new quorum, waiting" << dendl;
     }
@@ -650,11 +782,12 @@ void Monitor::slurp()
     }
 
     dout(10) << " " << p->first << " v " << p->second << " vs my " << pax->get_version() << dendl;
-    if (p->second > pax->get_version()) {
+    if (p->second > pax->get_version() ||
+	pax->get_stashed_version() > pax->get_version()) {
       if (!pax->is_slurping()) {
         pax->start_slurping();
       }
-      MMonProbe *m = new MMonProbe(monmap->fsid, MMonProbe::OP_SLURP, name);
+      MMonProbe *m = new MMonProbe(monmap->fsid, MMonProbe::OP_SLURP, name, has_ever_joined);
       m->machine_name = p->first;
       m->oldest_version = pax->get_first_committed();
       m->newest_version = pax->get_version();
@@ -668,7 +801,7 @@ void Monitor::slurp()
       if (!pax->is_slurping()) {
         pax->start_slurping();
       }
-      MMonProbe *m = new MMonProbe(monmap->fsid, MMonProbe::OP_SLURP_LATEST, name);
+      MMonProbe *m = new MMonProbe(monmap->fsid, MMonProbe::OP_SLURP_LATEST, name, has_ever_joined);
       m->machine_name = p->first;
       m->oldest_version = pax->get_first_committed();
       m->newest_version = pax->get_version();
@@ -691,7 +824,7 @@ void Monitor::slurp()
 
 MMonProbe *Monitor::fill_probe_data(MMonProbe *m, Paxos *pax)
 {
-  MMonProbe *r = new MMonProbe(monmap->fsid, MMonProbe::OP_DATA, name);
+  MMonProbe *r = new MMonProbe(monmap->fsid, MMonProbe::OP_DATA, name, has_ever_joined);
   r->machine_name = m->machine_name;
   r->oldest_version = pax->get_first_committed();
   r->newest_version = pax->get_version();
@@ -749,6 +882,10 @@ void Monitor::handle_probe_data(MMonProbe *m)
   if (m->oldest_version > pax->get_first_committed())
     pax->trim_to(m->oldest_version, true);
 
+  // note new latest version?
+  if (slurp_versions.count(m->machine_name))
+    slurp_versions[m->machine_name] = m->newest_version;
+
   // store any new stuff
   if (m->paxos_values.size()) {
     for (map<string, map<version_t, bufferlist> >::iterator p = m->paxos_values.begin();
@@ -791,7 +928,7 @@ void Monitor::win_standalone_election()
   assert(rank == 0);
   set<int> q;
   q.insert(rank);
-  win_election(1, q);
+  win_election(1, q, CEPH_FEATURES_ALL);
 }
 
 const utime_t& Monitor::get_leader_since() const
@@ -805,7 +942,7 @@ epoch_t Monitor::get_epoch()
   return elector.get_epoch();
 }
 
-void Monitor::win_election(epoch_t epoch, set<int>& active) 
+void Monitor::win_election(epoch_t epoch, set<int>& active, unsigned features) 
 {
   if (!is_electing())
     reset();
@@ -814,8 +951,11 @@ void Monitor::win_election(epoch_t epoch, set<int>& active)
   leader_since = ceph_clock_now(g_ceph_context);
   leader = rank;
   quorum = active;
+  quorum_features = features;
   outside_quorum.clear();
-  dout(10) << "win_election, epoch " << epoch << " quorum is " << quorum << dendl;
+  dout(10) << "win_election, epoch " << epoch << " quorum is " << quorum
+	   << " features are " << quorum_features
+	   << dendl;
 
   clog.info() << "mon." << name << "@" << rank
 		<< " won leader election with quorum " << quorum << "\n";
@@ -835,6 +975,7 @@ void Monitor::lose_election(epoch_t epoch, set<int> &q, int l)
   leader = l;
   quorum = q;
   outside_quorum.clear();
+  quorum_features = 0;
   dout(10) << "lose_election, epoch " << epoch << " leader is mon" << leader
 	   << " quorum is " << quorum << dendl;
   
@@ -855,16 +996,18 @@ void Monitor::finish_election()
   update_logger();
   register_cluster_logger();
 
-  // make note of the fact that i was, once, part of the quorum.
-  store->put_int(1, "joined");
+  // am i named properly?
+  string cur_name = monmap->get_name(messenger->get_myaddr());
+  if (cur_name != name) {
+    dout(10) << " renaming myself from " << cur_name << " -> " << name << dendl;
+    messenger->send_message(new MMonJoin(monmap->fsid, name, messenger->get_myaddr()),
+			    monmap->get_inst(*quorum.begin()));
+  }
 } 
 
 
 bool Monitor::_allowed_command(MonSession *s, const vector<string>& cmd)
 {
-  if (s->caps.check_privileges(PAXOS_MONMAP, MON_CAP_ALL))
-    return true;
-
   for (list<list<string> >::iterator p = s->caps.cmd_allow.begin();
        p != s->caps.cmd_allow.end();
        ++p) {
@@ -943,6 +1086,41 @@ void Monitor::_mon_status(ostream& ss)
   jf.flush(ss);
 }
 
+void Monitor::get_health(string& status, bufferlist *detailbl)
+{
+  list<pair<health_status_t,string> > summary;
+  list<pair<health_status_t,string> > detail;
+  for (vector<PaxosService*>::iterator p = paxos_service.begin();
+       p != paxos_service.end();
+       p++) {
+    PaxosService *s = *p;
+    s->get_health(summary, detailbl ? &detail : NULL);
+  }
+
+  stringstream ss;
+  health_status_t overall = HEALTH_OK;
+  if (!summary.empty()) {
+    ss << ' ';
+    while (!summary.empty()) {
+      if (overall > summary.front().first)
+	overall = summary.front().first;
+      ss << summary.front().second;
+      summary.pop_front();
+      if (!summary.empty())
+	ss << "; ";
+    }
+  }
+  stringstream fss;
+  fss << overall;
+  status = fss.str() + ss.str();
+
+  while (!detail.empty()) {
+    detailbl->append(detail.front().second);
+    detailbl->append('\n');
+    detail.pop_front();
+  }
+}
+
 void Monitor::handle_command(MMonCommand *m)
 {
   if (m->fsid != monmap->fsid) {
@@ -952,17 +1130,22 @@ void Monitor::handle_command(MMonCommand *m)
   }
 
   MonSession *session = m->get_session();
-  if (!session || !_allowed_command(session, m->cmd)) {
+  if (!session) {
     string rs = "Access denied";
     reply_command(m, -EACCES, rs, 0);
     return;
   }
 
+  bool access_cmd = _allowed_command(session, m->cmd);
+  bool access_r = (session->caps.check_privileges(PAXOS_MONMAP, MON_CAP_R) ||
+		   access_cmd);
+  bool access_all = (session->caps.get_allow_all() || access_cmd);
+
   dout(0) << "handle_command " << *m << dendl;
   bufferlist rdata;
   string rs;
   int r = -EINVAL;
-  rs = "unrecognized subsystem";
+  rs = "unrecognized command";
   if (!m->cmd.empty()) {
     if (m->cmd[0] == "mds") {
       mdsmon()->dispatch(m);
@@ -987,6 +1170,11 @@ void Monitor::handle_command(MMonCommand *m)
       return;
     }
     if (m->cmd[0] == "log") {
+      if (!access_r) {
+	r = -EACCES;
+	rs = "access denied";
+	goto out;
+      }
       stringstream ss;
       for (unsigned i=1; i<m->cmd.size(); i++) {
 	if (i > 1)
@@ -999,12 +1187,22 @@ void Monitor::handle_command(MMonCommand *m)
       return;
     }
     if (m->cmd[0] == "stop_cluster") {
+      if (!access_all) {
+	r = -EACCES;
+	rs = "access denied";
+	goto out;
+      }
       stop_cluster();
       reply_command(m, 0, "initiating cluster shutdown", 0);
       return;
     }
 
     if (m->cmd[0] == "injectargs") {
+      if (!access_all) {
+	r = -EACCES;
+	rs = "access denied";
+	goto out;
+      }
       if (m->cmd.size() == 2) {
 	dout(0) << "parsing injected options '" << m->cmd[1] << "'" << dendl;
 	ostringstream oss;
@@ -1026,10 +1224,34 @@ void Monitor::handle_command(MMonCommand *m)
       authmon()->dispatch(m);
       return;
     }
+    if (m->cmd[0] == "status") {
+      if (!access_r) {
+	r = -EACCES;
+	rs = "access denied";
+	goto out;
+      }
+      // reply with the status for all the components
+      string health;
+      get_health(health, NULL);
+      stringstream ss;
+      ss << "   health " << health << "\n";
+      ss << "   monmap " << *monmap << ", election epoch " << get_epoch() << ", quorum " << get_quorum()
+	 << " " << get_quorum_names() << "\n";
+      ss << "   osdmap " << osdmon()->osdmap << "\n";
+      ss << "    pgmap " << pgmon()->pg_map << "\n";
+      ss << "   mdsmap " << mdsmon()->mdsmap << "\n";
+      rs = ss.str();
+      r = 0;
+    }
     if (m->cmd[0] == "quorum_status") {
+      if (!access_r) {
+	r = -EACCES;
+	rs = "access denied";
+	goto out;
+      }
       // make sure our map is readable and up to date
       if (!is_leader() && !is_peon()) {
-	dout(10) << " waiting for qorum" << dendl;
+	dout(10) << " waiting for quorum" << dendl;
 	waitfor_quorum.push_back(new C_RetryMessage(this, m));
 	return;
       }
@@ -1039,52 +1261,42 @@ void Monitor::handle_command(MMonCommand *m)
       r = 0;
     }
     if (m->cmd[0] == "mon_status") {
+      if (!access_r) {
+	r = -EACCES;
+	rs = "access denied";
+	goto out;
+      }
       stringstream ss;
       _mon_status(ss);
       rs = ss.str();
       r = 0;
     }
     if (m->cmd[0] == "health") {
-      list<pair<health_status_t,string> > summary;
-      list<pair<health_status_t,string> > detail;
-      for (vector<PaxosService*>::iterator p = paxos_service.begin();
-	   p != paxos_service.end();
-	   p++) {
-	PaxosService *s = *p;
-	ostringstream oss;
-	s->get_health(summary, (m->cmd.size() > 1) ? &detail : NULL);
-      }
-      
-      stringstream ss;
-      health_status_t overall = HEALTH_OK;
-      if (!summary.empty()) {
-	ss << ' ';
-	while (!summary.empty()) {
-	  if (overall > summary.front().first)
-	    overall = summary.front().first;
-	  ss << summary.front().second;
-	  summary.pop_front();
-	  if (!summary.empty())
-	    ss << "; ";
-	}
-      }
-      stringstream fss;
-      fss << overall;
-      rs = fss.str() + ss.str();
-      while (!detail.empty()) {
-	rdata.append(detail.front().second);
-	rdata.append('\n');
-	detail.pop_front();
+      if (!access_r) {
+	r = -EACCES;
+	rs = "access denied";
+	goto out;
       }
+      get_health(rs, (m->cmd.size() > 1) ? &rdata : NULL);
       r = 0;
     }
     if (m->cmd[0] == "heap") {
+      if (!access_all) {
+	r = -EACCES;
+	rs = "access denied";
+	goto out;
+      }
       if (!ceph_using_tcmalloc())
 	rs = "tcmalloc not enabled, can't use heap profiler commands\n";
       else
 	ceph_heap_profiler_handle_command(m->cmd, clog);
     }
     if (m->cmd[0] == "quorum") {
+      if (!access_all) {
+	r = -EACCES;
+	rs = "access denied";
+	goto out;
+      }
       if (m->cmd[1] == "exit") {
         reset();
         start_election();
@@ -1102,9 +1314,9 @@ void Monitor::handle_command(MMonCommand *m)
 	r = -EINVAL;
       }
     }
-  } else 
-    rs = "no command";
+  }
 
+ out:
   if (!m->get_source().is_mon())  // don't reply to mon->mon commands
     reply_command(m, r, rs, rdata, 0);
   else
@@ -1147,7 +1359,7 @@ void Monitor::forward_request_leader(PaxosServiceMessage *req)
     RoutedRequest *rr = new RoutedRequest;
     rr->tid = ++routed_request_tid;
     rr->client = req->get_source_inst();
-    encode_message(req, -1, rr->request_bl);   // for my use only; use all features
+    encode_message(req, CEPH_FEATURES_ALL, rr->request_bl);   // for my use only; use all features
     rr->session = (MonSession *)session->get();
     routed_requests[rr->tid] = rr;
     session->routed_request_tids.insert(rr->tid);
@@ -1213,7 +1425,7 @@ void Monitor::try_send_message(Message *m, entity_inst_t to)
   dout(10) << "try_send_message " << *m << " to " << to << dendl;
 
   bufferlist bl;
-  encode_message(m, -1, bl);  // fixme: assume peers have all features we do.
+  encode_message(m, CEPH_FEATURES_ALL, bl);  // fixme: assume peers have all features we do.
 
   messenger->send_message(m, to);
 
@@ -1321,25 +1533,6 @@ void Monitor::remove_session(MonSession *s)
 }
 
 
-void Monitor::handle_observe(MMonObserve *m)
-{
-  dout(10) << "handle_observe " << *m << " from " << m->get_source_inst() << dendl;
-  // check that there are perms. Send a response back if they aren't sufficient,
-  // and delete the message (if it's not deleted for us, which happens when
-  // we own the connection to the requested observer).
-  MonSession *session = m->get_session();
-  if (!session || !session->caps.check_privileges(PAXOS_MONMAP, MON_CAP_X)) {
-    send_reply(m, m);
-    return;
-  }
-  if (m->machine_id >= PAXOS_NUM) {
-    dout(0) << "register_observer: bad monitor id: " << m->machine_id << dendl;
-  } else {
-    paxos[m->machine_id]->register_observer(m->get_orig_source_inst(), m->ver);
-  }
-  messenger->send_message(m, m->get_orig_source_inst());
-}
-
 void Monitor::send_command(const entity_inst_t& inst,
 			   const vector<string>& com, version_t version)
 {
@@ -1517,6 +1710,10 @@ bool Monitor::_ms_dispatch(Message *m)
       paxos_service[PAXOS_LOG]->dispatch((PaxosServiceMessage*)m);
       break;
 
+    case MSG_LOGACK:
+      clog.handle_log_ack((MLogAck*)m);
+      break;
+
       // monmap
     case MSG_MON_JOIN:
       paxos_service[PAXOS_MONMAP]->dispatch((PaxosServiceMessage*)m);
@@ -1556,10 +1753,6 @@ bool Monitor::_ms_dispatch(Message *m)
       }
       break;
 
-    case MSG_MON_OBSERVE:
-      handle_observe((MMonObserve *)m);
-      break;
-
       // elector messages
     case MSG_MON_ELECTION:
       //check privileges here for simplicity
@@ -1568,10 +1761,11 @@ bool Monitor::_ms_dispatch(Message *m)
 	dout(0) << "MMonElection received from entity without enough caps!"
 		<< s->caps << dendl;
       }
-      if (!is_probing() && !is_slurping())
+      if (!is_probing() && !is_slurping()) {
 	elector.dispatch(m);
-      else
+      } else {
 	m->put();
+      }
       break;
 
     case MSG_FORWARD:
@@ -1623,8 +1817,14 @@ void Monitor::handle_subscribe(MMonSubscribe *m)
       if ((int)s->caps.check_privileges(PAXOS_OSDMAP, MON_CAP_R)) {
         osdmon()->check_sub(s->sub_map["osdmap"]);
       }
+    } else if (p->first == "osd_pg_creates") {
+      if ((int)s->caps.check_privileges(PAXOS_OSDMAP, MON_CAP_W)) {
+	pgmon()->check_sub(s->sub_map["osd_pg_creates"]);
+      }
     } else if (p->first == "monmap") {
       check_sub(s->sub_map["monmap"]);
+    } else if (logmon()->sub_name_to_id(p->first) >= 0) {
+      logmon()->check_sub(s->sub_map[p->first]);
     }
   }
 
@@ -1728,10 +1928,7 @@ void Monitor::check_sub(Subscription *sub)
 void Monitor::send_latest_monmap(Connection *con)
 {
   bufferlist bl;
-  if (!con->has_feature(CEPH_FEATURE_MONNAMES))
-    monmap->encode_v1(bl);
-  else
-    monmap->encode(bl);
+  monmap->encode(bl, con->get_features());
   messenger->send_message(new MMonMap(bl), con);
 }
 
@@ -1770,8 +1967,11 @@ void Monitor::tick()
   // ok go.
   dout(11) << "tick" << dendl;
   
-  for (vector<PaxosService*>::iterator p = paxos_service.begin(); p != paxos_service.end(); p++)
-    (*p)->tick();
+  if (!is_slurping()) {
+    for (vector<PaxosService*>::iterator p = paxos_service.begin(); p != paxos_service.end(); p++) {
+      (*p)->tick();
+    }
+  }
   
   // trim sessions
   utime_t now = ceph_clock_now(g_ceph_context);
@@ -1834,7 +2034,7 @@ int Monitor::mkfs(bufferlist& osdmapbl)
 
   // save monmap, osdmap, keyring.
   bufferlist monmapbl;
-  monmap->encode(monmapbl);
+  monmap->encode(monmapbl, CEPH_FEATURES_ALL);
   monmap->set_epoch(0);     // must be 0 to avoid confusing first MonmapMonitor::update_from_paxos()
   store->put_bl_ss(monmapbl, "mkfs", "monmap");
 
@@ -1857,13 +2057,33 @@ int Monitor::mkfs(bufferlist& osdmapbl)
     derr << "unable to load initial keyring " << g_conf->keyring << dendl;
     return r;
   }
+
+  // put mon. key in external keyring; seed with everything else.
+  extract_save_mon_key(keyring);
+
   bufferlist keyringbl;
-  ::encode(keyring, keyringbl);
+  keyring.encode_plaintext(keyringbl);
   store->put_bl_ss(keyringbl, "mkfs", "keyring");
 
   return 0;
 }
 
+void Monitor::extract_save_mon_key(KeyRing& keyring)
+{
+  EntityName mon_name;
+  mon_name.set_type(CEPH_ENTITY_TYPE_MON);
+  EntityAuth mon_key;
+  if (keyring.get_auth(mon_name, mon_key)) {
+    dout(10) << "extract_save_mon_key moving mon. key to separate keyring" << dendl;
+    KeyRing pkey;
+    pkey.add(mon_name, mon_key);
+    bufferlist bl;
+    pkey.encode_plaintext(bl);
+    store->put_bl_ss(bl, "keyring", NULL);
+    keyring.remove(mon_name);
+  }
+}
+
 bool Monitor::ms_get_authorizer(int service_id, AuthAuthorizer **authorizer, bool force_new)
 {
   dout(10) << "ms_get_authorizer for " << ceph_entity_type_name(service_id) << dendl;
@@ -1888,8 +2108,9 @@ bool Monitor::ms_get_authorizer(int service_id, AuthAuthorizer **authorizer, boo
   auth_ticket_info.ticket.global_id = 0;
 
   CryptoKey secret;
-  if (!key_server.get_secret(name, secret)) {
-    dout(0) << " couldn't get secret for mon service" << dendl;
+  if (!keyring.get_secret(name, secret) &&
+      !key_server.get_secret(name, secret)) {
+    dout(0) << " couldn't get secret for mon service from keyring or keyserver" << dendl;
     stringstream ss;
     key_server.list_secrets(ss);
     dout(0) << ss.str() << dendl;
@@ -1943,7 +2164,7 @@ bool Monitor::ms_verify_authorizer(Connection *con, int peer_type,
       CephXServiceTicketInfo auth_ticket_info;
       
       if (authorizer_data.length()) {
-	int ret = cephx_verify_authorizer(g_ceph_context, &key_server, iter,
+	int ret = cephx_verify_authorizer(g_ceph_context, &keyring, iter,
 					  auth_ticket_info, authorizer_reply);
 	if (ret >= 0)
 	  isvalid = true;
