@@ -76,6 +76,9 @@ void PGMonitor::on_active()
   }
 
   update_logger();
+
+  if (mon->is_leader())
+    mon->clog.info() << "pgmap " << pg_map << "\n";
 }
 
 void PGMonitor::update_logger()
@@ -93,13 +96,13 @@ void PGMonitor::update_logger()
   for (hash_map<int,int>::iterator p = pg_map.num_pg_by_state.begin();
        p != pg_map.num_pg_by_state.end();
        ++p) {
-    if (p->second & PG_STATE_ACTIVE) {
-      active++;
-      if (p->second & PG_STATE_CLEAN)
-	active_clean++;
+    if (p->first & PG_STATE_ACTIVE) {
+      active += p->second;
+      if (p->first & PG_STATE_CLEAN)
+	active_clean += p->second;
     }
-    if (p->second & PG_STATE_PEERING)
-      peering++;
+    if (p->first & PG_STATE_PEERING)
+      peering += p->second;
   }
   mon->cluster_logger->set(l_cluster_num_pg_active_clean, active_clean);
   mon->cluster_logger->set(l_cluster_num_pg_active, active);
@@ -194,7 +197,7 @@ void PGMonitor::update_from_paxos()
 
   // save latest
   bufferlist bl;
-  pg_map.encode(bl);
+  pg_map.encode(bl, mon->get_quorum_features());
   paxos->stash_latest(paxosv, bl);
 
   // dump pgmap summaries?  (useful for debugging)
@@ -254,7 +257,7 @@ void PGMonitor::encode_pending(bufferlist &bl)
 {
   dout(10) << "encode_pending v " << pending_inc.version << dendl;
   assert(paxos->get_version() + 1 == pending_inc.version);
-  pending_inc.encode(bl);
+  pending_inc.encode(bl, mon->get_quorum_features());
 }
 
 bool PGMonitor::preprocess_query(PaxosServiceMessage *m)
@@ -377,7 +380,7 @@ bool PGMonitor::preprocess_pg_stats(MPGStats *stats)
   // check caps
   MonSession *session = stats->get_session();
   if (!session) {
-    derr << "PGMonitor::preprocess_pg_stats: no monitor session!" << dendl;
+    dout(10) << "PGMonitor::preprocess_pg_stats: no monitor session!" << dendl;
     stats->put();
     return true;
   }
@@ -594,12 +597,24 @@ void PGMonitor::check_osd_map(epoch_t epoch)
 	 ++p) {
       if (p->second & CEPH_OSD_UP) {   // true if marked up OR down, but we're too lazy to check which
 	need_check_down_pgs = true;
+
 	// clear out the last_osd_report for this OSD
         map<int, utime_t>::iterator report = last_osd_report.find(p->first);
         if (report != last_osd_report.end()) {
           last_osd_report.erase(report);
         }
       }
+
+      if (p->second & CEPH_OSD_EXISTS) {
+	// whether it was created *or* destroyed, we can safely drop
+	// it's osd_stat_t record.
+	dout(10) << "check_osd_map  osd." << p->first << " created or destroyed" << dendl;
+	pending_inc.osd_stat_rm.insert(p->first);
+
+	// and adjust full, nearfull set
+	pg_map.nearfull_osds.erase(p->first);
+	pg_map.full_osds.erase(p->first);
+      }
     }
   }
 
@@ -744,7 +759,6 @@ void PGMonitor::send_pg_creates()
 {
   dout(10) << "send_pg_creates to " << pg_map.creating_pgs.size() << " pgs" << dendl;
 
-  map<int, MOSDPGCreate*> msg;
   utime_t now = ceph_clock_now(g_ceph_context);
   
   for (set<pg_t>::iterator p = pg_map.creating_pgs.begin();
@@ -752,42 +766,65 @@ void PGMonitor::send_pg_creates()
        p++) {
     pg_t pgid = *p;
     pg_t on = pgid;
-    if (pg_map.pg_stat[pgid].parent_split_bits)
-      on = pg_map.pg_stat[pgid].parent;
+    pg_stat_t& s = pg_map.pg_stat[pgid];
+    if (s.parent_split_bits)
+      on = s.parent;
     vector<int> acting;
     int nrep = mon->osdmon()->osdmap.pg_to_acting_osds(on, acting);
-    if (!nrep) {
-      dout(20) << "send_pg_creates  " << pgid << " -> no osds in epoch "
-	       << mon->osdmon()->osdmap.get_epoch() << ", skipping" << dendl;
-      continue;  // blarney!
-    }
-    int osd = acting[0];
+
+    if (s.acting.size())
+      pg_map.creating_pgs_by_osd[s.acting[0]].erase(pgid);
+    s.acting = acting;
 
     // don't send creates for localized pgs
     if (pgid.preferred() >= 0)
       continue;
 
+    if (nrep) {
+      pg_map.creating_pgs_by_osd[acting[0]].insert(pgid);
+    } else {
+      dout(20) << "send_pg_creates  " << pgid << " -> no osds in epoch "
+	       << mon->osdmon()->osdmap.get_epoch() << ", skipping" << dendl;
+      continue;  // blarney!
+    }
+  }
+
+  for (map<int, set<pg_t> >::iterator p = pg_map.creating_pgs_by_osd.begin();
+       p != pg_map.creating_pgs_by_osd.end();
+       ++p) {
+    int osd = p->first;
+
     // throttle?
     if (last_sent_pg_create.count(osd) &&
 	now - g_conf->mon_pg_create_interval < last_sent_pg_create[osd]) 
       continue;
       
-    dout(20) << "send_pg_creates  " << pgid << " -> osd." << osd 
-	     << " in epoch " << pg_map.pg_stat[pgid].created << dendl;
-    if (msg.count(osd) == 0)
-      msg[osd] = new MOSDPGCreate(mon->osdmon()->osdmap.get_epoch());
-    msg[osd]->mkpg[pgid] = pg_create_t(pg_map.pg_stat[pgid].created,
-				       pg_map.pg_stat[pgid].parent,
-				       pg_map.pg_stat[pgid].parent_split_bits);
+    if (mon->osdmon()->osdmap.is_up(osd))
+      send_pg_creates(osd, NULL);
+  }
+}
+
+void PGMonitor::send_pg_creates(int osd, Connection *con)
+{
+  map<int, set<pg_t> >::iterator p = pg_map.creating_pgs_by_osd.find(osd);
+  if (p == pg_map.creating_pgs_by_osd.end())
+    return;
+
+  dout(20) << "send_pg_creates osd." << osd << " pgs " << p->second << dendl;
+  MOSDPGCreate *m = new MOSDPGCreate(mon->osdmon()->osdmap.get_epoch());
+  for (set<pg_t>::iterator q = p->second.begin(); q != p->second.end(); ++q) {
+    m->mkpg[*q] = pg_create_t(pg_map.pg_stat[*q].created,
+			      pg_map.pg_stat[*q].parent,
+			      pg_map.pg_stat[*q].parent_split_bits);
   }
 
-  for (map<int, MOSDPGCreate*>::iterator p = msg.begin();
-       p != msg.end();
-       p++) {
-    dout(10) << "sending pg_create to osd." << p->first << dendl;
-    mon->messenger->send_message(p->second, mon->osdmon()->osdmap.get_inst(p->first));
-    last_sent_pg_create[p->first] = ceph_clock_now(g_ceph_context);
+  if (con) {
+    mon->messenger->send_message(m, con);
+  } else {
+    assert(mon->osdmon()->osdmap.is_up(osd));
+    mon->messenger->send_message(m, mon->osdmon()->osdmap.get_inst(osd));
   }
+  last_sent_pg_create[osd] = ceph_clock_now(g_ceph_context);
 }
 
 bool PGMonitor::check_down_pgs()
@@ -829,6 +866,15 @@ bool PGMonitor::preprocess_command(MMonCommand *m)
   bufferlist rdata;
   stringstream ss;
 
+  MonSession *session = m->get_session();
+  if (!session ||
+      (!session->caps.get_allow_all() &&
+       !session->caps.check_privileges(PAXOS_PGMAP, MON_CAP_R) &&
+       !mon->_allowed_command(session, m->cmd))) {
+    mon->reply_command(m, -EACCES, "access denied", rdata, paxos->get_version());
+    return true;
+  }
+
   vector<const char*> args;
   for (unsigned i = 1; i < m->cmd.size(); i++)
     args.push_back(m->cmd[i].c_str());
@@ -1027,6 +1073,15 @@ bool PGMonitor::prepare_command(MMonCommand *m)
   int r = -EINVAL;
   string rs;
 
+  MonSession *session = m->get_session();
+  if (!session ||
+      (!session->caps.get_allow_all() &&
+       !session->caps.check_privileges(PAXOS_PGMAP, MON_CAP_W) &&
+       !mon->_allowed_command(session, m->cmd))) {
+    mon->reply_command(m, -EACCES, "access denied", paxos->get_version());
+    return true;
+  }
+
   if (m->cmd.size() >= 1 && m->cmd[1] == "force_create_pg") {
     if (m->cmd.size() <= 2) {
       ss << "usage: pg force_create_pg <pg>";
@@ -1096,6 +1151,19 @@ bool PGMonitor::prepare_command(MMonCommand *m)
   return false;
 }
 
+static void note_stuck_detail(hash_map<pg_t,pg_stat_t>& stuck_pgs,
+			      list<pair<health_status_t,string> > *detail)
+{
+  for (hash_map<pg_t,pg_stat_t>::iterator p = stuck_pgs.begin();
+       p != stuck_pgs.end();
+       ++p) {
+    ostringstream ss;
+    ss << "pg " << p->first << " is stuck " << pg_state_string(p->second.state)
+       << ", last acting " << p->second.acting;
+    detail->push_back(make_pair(HEALTH_WARN, ss.str()));
+  }
+}
+
 void PGMonitor::get_health(list<pair<health_status_t,string> >& summary,
 			   list<pair<health_status_t,string> > *detail) const
 {
@@ -1132,29 +1200,24 @@ void PGMonitor::get_health(list<pair<health_status_t,string> >& summary,
   pg_map.get_stuck_stats(PGMap::STUCK_INACTIVE, cutoff, stuck_pgs);
   if (!stuck_pgs.empty()) {
     note["stuck inactive"] = stuck_pgs.size();
+    if (detail)
+      note_stuck_detail(stuck_pgs, detail);
   }
   stuck_pgs.clear();
 
   pg_map.get_stuck_stats(PGMap::STUCK_UNCLEAN, cutoff, stuck_pgs);
   if (!stuck_pgs.empty()) {
     note["stuck unclean"] = stuck_pgs.size();
+    if (detail)
+      note_stuck_detail(stuck_pgs, detail);
   }
   stuck_pgs.clear();
 
   pg_map.get_stuck_stats(PGMap::STUCK_STALE, cutoff, stuck_pgs);
   if (!stuck_pgs.empty()) {
     note["stuck stale"] = stuck_pgs.size();
-  }
-
-  if (detail) {
-    for (hash_map<pg_t,pg_stat_t>::iterator p = stuck_pgs.begin();
-	 p != stuck_pgs.end();
-	 ++p) {
-      ostringstream ss;
-      ss << "pg " << p->first << " is stuck " << pg_state_string(p->second.state)
-	 << ", last acting " << p->second.acting;
-      detail->push_back(make_pair(HEALTH_WARN, ss.str()));
-    }
+    if (detail)
+      note_stuck_detail(stuck_pgs, detail);
   }
 
   if (!note.empty()) {
@@ -1180,6 +1243,7 @@ void PGMonitor::get_health(list<pair<health_status_t,string> >& summary,
 	    stuck_pgs.count(p->first) == 0) {
 	  ostringstream ss;
 	  ss << "pg " << p->first << " is " << pg_state_string(p->second.state);
+	  ss << ", acting " << p->second.acting;
 	  if (p->second.stats.sum.num_objects_unfound)
 	    ss << ", " << p->second.stats.sum.num_objects_unfound << " unfound";
 	  detail->push_back(make_pair(HEALTH_WARN, ss.str()));
@@ -1301,3 +1365,11 @@ int PGMonitor::dump_stuck_pg_stats(ostream& ss,
   ss << "ok";
   return 0;
 }
+
+void PGMonitor::check_sub(Subscription *sub)
+{
+  if (sub->type == "osd_pg_creates") {
+    send_pg_creates(sub->session->inst.name.num(),
+		    sub->session->con);
+  }
+}
