@@ -43,10 +43,6 @@ class PGMonitor : public PaxosService {
 public:
   PGMap pg_map;
 
-  Mutex ratio_lock;
-  bool need_full_ratio_update, need_nearfull_ratio_update;
-  float new_full_ratio, new_nearfull_ratio;
-
   bool need_check_down_pgs;
 
 private:
@@ -68,8 +64,6 @@ private:
   bool prepare_pg_stats(MPGStats *stats);
   void _updated_stats(MPGStats *req, MPGStatsAck *ack);
 
-  void update_full_ratios(float full_ratio, float nearfull_ratio);
-
   struct C_Stats : public Context {
     PGMonitor *pgmon;
     MPGStats *req;
@@ -102,6 +96,7 @@ private:
   bool register_new_pgs();
 
   void send_pg_creates();
+  void send_pg_creates(int osd, Connection *con);
 
   /**
    * check pgs for down primary osds
@@ -136,15 +131,18 @@ public:
 
   void check_osd_map(epoch_t epoch);
 
-  enum health_status_t get_health(std::ostream &ss) const;
+  void get_health(list<pair<health_status_t,string> >& summary,
+		  list<pair<health_status_t,string> > *detail) const;
+  void check_full_osd_health(list<pair<health_status_t,string> >& summary,
+			     list<pair<health_status_t,string> > *detail,
+			     const set<int>& s, const char *desc, health_status_t sev) const;
+
+  void check_sub(Subscription *sub);
 
 private:
   // no copying allowed
   PGMonitor(const PGMonitor &rhs);
   PGMonitor &operator=(const PGMonitor &rhs);
-
-  RatioMonitor *ratio_monitor;
-  friend class RatioMonitor;
 };
 
 #endif
