@@ -15,8 +15,17 @@
 #ifndef CEPH_MESSAGE_H
 #define CEPH_MESSAGE_H
  
-/* public message types */
+#include <stdlib.h>
+#include <ostream>
+
 #include "include/types.h"
+#include "include/buffer.h"
+#include "common/Throttle.h"
+#include "msg_types.h"
+
+#include "common/RefCountedObj.h"
+
+#include "common/debug.h"
 #include "common/config.h"
 
 // monitor internal
@@ -30,8 +39,8 @@
 #define MSG_MON_COMMAND_ACK        51
 #define MSG_LOG                    52
 #define MSG_LOGACK                 53
-#define MSG_MON_OBSERVE            54
-#define MSG_MON_OBSERVE_NOTIFY     55
+//#define MSG_MON_OBSERVE            54
+//#define MSG_MON_OBSERVE_NOTIFY     55
 #define MSG_CLASS                  56
 #define MSG_CLASS_ACK              57
 
@@ -128,45 +137,12 @@
 
 
 
-#include <stdlib.h>
-
-#include <iostream>
-#include <list>
-using std::list;
-
-#include <ext/hash_map>
-
-
-#include "include/types.h"
-#include "include/buffer.h"
-#include "common/Throttle.h"
-#include "msg_types.h"
-#include "include/atomic.h"
-
-#include "common/debug.h"
-
 
 
 // ======================================================
 
 // abstract Connection, for keeping per-connection state
 
-struct RefCountedObject {
-  atomic_t nref;
-  RefCountedObject() : nref(1) {}
-  virtual ~RefCountedObject() {}
-  
-  RefCountedObject *get() {
-    //generic_dout(0) << "RefCountedObject::get " << this << " " << nref.read() << " -> " << (nref.read() + 1) << dendl;
-    nref.inc();
-    return this;
-  }
-  void put() {
-    //generic_dout(0) << "RefCountedObject::put " << this << " " << nref.read() << " -> " << (nref.read() - 1) << dendl;
-    if (nref.dec() == 0)
-      delete this;
-  }
-};
 
 struct Connection : public RefCountedObject {
   Mutex lock;
@@ -215,9 +191,9 @@ public:
       return pipe->get();
     return NULL;
   }
-  void clear_pipe() {
-    Mutex::Locker l(lock);
-    if (pipe) {
+  void clear_pipe(RefCountedObject *old_p) {
+    if(old_p == pipe) {
+      Mutex::Locker l(lock);
       pipe->put();
       pipe = NULL;
     }
@@ -478,4 +454,10 @@ inline ostream& operator<<(ostream& out, Message& m) {
 extern void encode_message(Message *m, uint64_t features, bufferlist& bl);
 extern Message *decode_message(CephContext *cct, bufferlist::iterator& bl);
 
+namespace boost {
+  void intrusive_ptr_add_ref(Message *p);
+  void intrusive_ptr_release(Message *p);
+};
+
+
 #endif
