@@ -28,6 +28,8 @@
 #include "common/Mutex.h"
 #include "common/Clock.h"
 
+#include "include/ceph_features.h"
+
 #include "crush/CrushWrapper.h"
 
 #include "include/interval_set.h"
@@ -42,21 +44,6 @@ using namespace std;
 #include <ext/hash_set>
 using __gnu_cxx::hash_set;
 
-
-
-/*
- * some system constants
- */
-
-// pg roles
-#define PG_ROLE_STRAY   -1
-#define PG_ROLE_HEAD     0
-#define PG_ROLE_ACKER    1
-#define PG_ROLE_MIDDLE   2  // der.. misnomer
-//#define PG_ROLE_TAIL     2
-
-
-
 /*
  * we track up to two intervals during which the osd was alive and
  * healthy.  the most recent is [up_from,up_thru), where up_thru is
@@ -130,6 +117,7 @@ public:
     map<int32_t,epoch_t> new_up_thru;
     map<int32_t,pair<epoch_t,epoch_t> > new_last_clean_interval;
     map<int32_t,epoch_t> new_lost;
+    map<int32_t,uuid_d> new_uuid;
 
     map<entity_addr_t,utime_t> new_blacklist;
     vector<entity_addr_t> old_blacklist;
@@ -137,8 +125,12 @@ public:
 
     string cluster_snapshot;
 
+    int get_net_marked_out(const OSDMap *previous) const;
+    int get_net_marked_down(const OSDMap *previous) const;
+    int identify_osd(uuid_d u) const;
+
     void encode_client_old(bufferlist& bl) const;
-    void encode(bufferlist& bl, uint64_t features=-1) const;
+    void encode(bufferlist& bl, uint64_t features=CEPH_FEATURES_ALL) const;
     void decode(bufferlist::iterator &p);
     void dump(Formatter *f) const;
     static void generate_test_instances(list<Incremental*>& o);
@@ -167,24 +159,32 @@ private:
   int num_osd;         // not saved
   int32_t max_osd;
   vector<uint8_t> osd_state;
-  vector<entity_addr_t> osd_addr;
-  vector<entity_addr_t> osd_cluster_addr;
-  vector<entity_addr_t> osd_hb_addr;
+
+  struct addrs_s {
+    vector<std::tr1::shared_ptr<entity_addr_t> > client_addr;
+    vector<std::tr1::shared_ptr<entity_addr_t> > cluster_addr;
+    vector<std::tr1::shared_ptr<entity_addr_t> > hb_addr;
+    entity_addr_t blank;
+  };
+  std::tr1::shared_ptr<addrs_s> osd_addrs;
+
   vector<__u32>   osd_weight;   // 16.16 fixed point, 0x10000 = "in", 0 = "out"
   vector<osd_info_t> osd_info;
-  map<pg_t,vector<int> > pg_temp;  // temp pg mapping (e.g. while we rebuild)
+  std::tr1::shared_ptr< map<pg_t,vector<int> > > pg_temp;  // temp pg mapping (e.g. while we rebuild)
 
   map<int64_t,pg_pool_t> pools;
   map<int64_t,string> pool_name;
   map<string,int64_t> name_pool;
 
+  std::tr1::shared_ptr< vector<uuid_d> > osd_uuid;
+
   hash_map<entity_addr_t,utime_t> blacklist;
 
   epoch_t cluster_snapshot_epoch;
   string cluster_snapshot;
 
  public:
-  CrushWrapper     crush;       // hierarchical map
+  std::tr1::shared_ptr<CrushWrapper> crush;       // hierarchical map
 
   friend class OSDMonitor;
   friend class PGMonitor;
@@ -195,7 +195,11 @@ private:
 	     pool_max(-1),
 	     flags(0),
 	     num_osd(0), max_osd(0),
-	     cluster_snapshot_epoch(0) { 
+	     osd_addrs(new addrs_s),
+	     pg_temp(new map<pg_t,vector<int> >),
+	     osd_uuid(new vector<uuid_d>),
+	     cluster_snapshot_epoch(0),
+	     crush(new CrushWrapper) {
     memset(&fsid, 0, sizeof(fsid));
   }
 
@@ -206,13 +210,7 @@ private:
   epoch_t get_epoch() const { return epoch; }
   void inc_epoch() { epoch++; }
 
-  void set_epoch(epoch_t e) {
-    epoch = e;
-    for (map<int64_t,pg_pool_t>::iterator p = pools.begin();
-	 p != pools.end();
-	 p++)
-      p->second.last_change = e;
-  }
+  void set_epoch(epoch_t e);
 
   /* stamps etc */
   const utime_t& get_created() const { return created; }
@@ -236,40 +234,17 @@ private:
   }
   int calc_num_osds();
 
-  void get_all_osds(set<int32_t>& ls) const {
-    for (int i=0; i<max_osd; i++)
-      if (exists(i))
-	ls.insert(i);
-  }
-  int get_num_up_osds() const {
-    int n = 0;
-    for (int i=0; i<max_osd; i++)
-      if (osd_state[i] & CEPH_OSD_EXISTS &&
-	  osd_state[i] & CEPH_OSD_UP) n++;
-    return n;
-  }
-  int get_num_in_osds() const {
-    int n = 0;
-    for (int i=0; i<max_osd; i++)
-      if (osd_state[i] & CEPH_OSD_EXISTS &&
-	  get_weight(i) != CEPH_OSD_OUT) n++;
-    return n;
-  }
+  void get_all_osds(set<int32_t>& ls) const;
+  int get_num_up_osds() const;
+  int get_num_in_osds() const;
 
   int get_flags() const { return flags; }
   int test_flag(int f) const { return flags & f; }
   void set_flag(int f) { flags |= f; }
   void clear_flag(int f) { flags &= ~f; }
 
-  static void calc_state_set(int state, set<string>& st) {
-    unsigned t = state;
-    for (unsigned s = 1; t; s <<= 1) {
-      if (t & s) {
-	t &= ~s;
-	st.insert(ceph_osd_state_name(s));
-      }
-    }
-  }
+  static void calc_state_set(int state, set<string>& st);
+
   int get_state(int o) const {
     assert(o < max_osd);
     return osd_state[o];
@@ -323,54 +298,46 @@ private:
     return exists(osd) && !is_out(osd);
   }
   
-  int identify_osd(const entity_addr_t& addr) const {
-    for (unsigned i=0; i<osd_addr.size(); i++)
-      if ((osd_addr[i] == addr) || (osd_cluster_addr[i] == addr))
-	return i;
-    return -1;
-  }
+  int identify_osd(const entity_addr_t& addr) const;
+  int identify_osd(const uuid_d& u) const;
+
   bool have_addr(const entity_addr_t& addr) const {
     return identify_osd(addr) >= 0;
   }
-  bool find_osd_on_ip(const entity_addr_t& ip) const {
-    for (unsigned i=0; i<osd_addr.size(); i++)
-      if (osd_addr[i].is_same_host(ip) || osd_cluster_addr[i].is_same_host(ip))
-	return i;
-    return -1;
-  }
+  bool find_osd_on_ip(const entity_addr_t& ip) const;
   bool have_inst(int osd) const {
     return exists(osd) && is_up(osd); 
   }
   const entity_addr_t &get_addr(int osd) const {
     assert(exists(osd));
-    return osd_addr[osd];
+    return osd_addrs->client_addr[osd] ? *osd_addrs->client_addr[osd] : osd_addrs->blank;
   }
   const entity_addr_t &get_cluster_addr(int osd) const {
     assert(exists(osd));
-    if (osd_cluster_addr[osd] == entity_addr_t())
+    if (!osd_addrs->cluster_addr[osd] || *osd_addrs->cluster_addr[osd] == entity_addr_t())
       return get_addr(osd);
-    return osd_cluster_addr[osd];
+    return *osd_addrs->cluster_addr[osd];
   }
   const entity_addr_t &get_hb_addr(int osd) const {
     assert(exists(osd));
-    return osd_hb_addr[osd];
+    return osd_addrs->hb_addr[osd] ? *osd_addrs->hb_addr[osd] : osd_addrs->blank;
   }
   entity_inst_t get_inst(int osd) const {
-    assert(exists(osd));
     assert(is_up(osd));
-    return entity_inst_t(entity_name_t::OSD(osd), osd_addr[osd]);
+    return entity_inst_t(entity_name_t::OSD(osd), get_addr(osd));
   }
   entity_inst_t get_cluster_inst(int osd) const {
-    assert(exists(osd));
     assert(is_up(osd));
-    if (osd_cluster_addr[osd] == entity_addr_t())
-      return get_inst(osd);
-    return entity_inst_t(entity_name_t::OSD(osd), osd_cluster_addr[osd]);
+    return entity_inst_t(entity_name_t::OSD(osd), get_cluster_addr(osd));
   }
   entity_inst_t get_hb_inst(int osd) const {
-    assert(exists(osd));
     assert(is_up(osd));
-    return entity_inst_t(entity_name_t::OSD(osd), osd_hb_addr[osd]);
+    return entity_inst_t(entity_name_t::OSD(osd), get_hb_addr(osd));
+  }
+
+  const uuid_d& get_uuid(int osd) const {
+    assert(exists(osd));
+    return (*osd_uuid)[osd];
   }
 
   const epoch_t& get_up_from(int osd) const {
@@ -399,34 +366,20 @@ private:
 
   int apply_incremental(Incremental &inc);
 
+  /// try to re-use/reference addrs in oldmap from newmap
+  static void dedup(const OSDMap *oldmap, OSDMap *newmap);
+
   // serialize, unserialize
 private:
   void encode_client_old(bufferlist& bl) const;
 public:
-  void encode(bufferlist& bl, uint64_t features=-1) const;
+  void encode(bufferlist& bl, uint64_t features=CEPH_FEATURES_ALL) const;
   void decode(bufferlist& bl);
   void decode(bufferlist::iterator& p);
 
 
   /****   mapping facilities   ****/
-  int object_locator_to_pg(const object_t& oid, const object_locator_t& loc, pg_t &pg) const {
-    // calculate ps (placement seed)
-    const pg_pool_t *pool = get_pg_pool(loc.get_pool());
-    if (!pool)
-      return -ENOENT;
-    ps_t ps;
-    if (loc.key.length())
-      ps = ceph_str_hash(pool->object_hash, loc.key.c_str(), loc.key.length());
-    else
-      ps = ceph_str_hash(pool->object_hash, oid.name.c_str(), oid.name.length());
-    // mix in preferred osd, so we don't get the same peers for
-    // all of the placement pgs (e.g. 0.0p*)
-    if (loc.get_preferred() >= 0)
-      ps += loc.get_preferred();
-    pg = pg_t(ps, loc.get_pool(), loc.get_preferred());
-    return 0;
-  }
-
+  int object_locator_to_pg(const object_t& oid, const object_locator_t& loc, pg_t &pg) const;
   pg_t object_locator_to_pg(const object_t& oid, const object_locator_t& loc) const {
     pg_t pg;
     int ret = object_locator_to_pg(oid, loc, pg);
@@ -435,25 +388,15 @@ public:
   }
 
   static object_locator_t file_to_object_locator(const ceph_file_layout& layout) {
-    return object_locator_t(layout.fl_pg_pool, layout.fl_pg_preferred);
+    return object_locator_t(layout.fl_pg_pool);
   }
 
   // oid -> pg
   ceph_object_layout file_to_object_layout(object_t oid, ceph_file_layout& layout) const {
-    return make_object_layout(oid, layout.fl_pg_pool,
-			      layout.fl_pg_preferred);
+    return make_object_layout(oid, layout.fl_pg_pool);
   }
 
-  ceph_object_layout make_object_layout(object_t oid, int pg_pool, int preferred=-1) const {
-    object_locator_t loc(pg_pool);
-    loc.preferred = preferred;
-    
-    ceph_object_layout ol;
-    pg_t pgid = object_locator_to_pg(oid, loc);
-    ol.ol_pgid = pgid.get_old_pg().v;
-    ol.ol_stripe_unit = 0;
-    return ol;
-  }
+  ceph_object_layout make_object_layout(object_t oid, int pg_pool) const;
 
   int get_pg_num(int pg_pool) const
   {
@@ -461,91 +404,21 @@ public:
     return pool->get_pg_num();
   }
 
-  // pg -> (osd list)
 private:
-  int _pg_to_osds(const pg_pool_t& pool, pg_t pg, vector<int>& osds) const {
-    // map to osds[]
-    ps_t pps = pool.raw_pg_to_pps(pg);  // placement ps
-    unsigned size = pool.get_size();
-    {
-      int preferred = pg.preferred();
-      if (preferred >= max_osd || preferred >= crush.get_max_devices())
-	preferred = -1;
+  /// pg -> (raw osd list)
+  int _pg_to_osds(const pg_pool_t& pool, pg_t pg, vector<int>& osds) const;
+  void _remove_nonexistent_osds(vector<int>& osds) const;
 
-      assert(get_max_osd() >= crush.get_max_devices());
+  /// pg -> (up osd list)
+  void _raw_to_up_osds(pg_t pg, vector<int>& raw, vector<int>& up) const;
 
-      // what crush rule?
-      int ruleno = crush.find_rule(pool.get_crush_ruleset(), pool.get_type(), size);
-      if (ruleno >= 0)
-	crush.do_rule(ruleno, pps, osds, size, preferred, osd_weight);
-    }
-  
-    return osds.size();
-  }
-
-  // pg -> (up osd list)
-  void _raw_to_up_osds(pg_t pg, vector<int>& raw, vector<int>& up) const {
-    up.clear();
-    for (unsigned i=0; i<raw.size(); i++) {
-      if (!exists(raw[i]) || is_down(raw[i])) 
-	continue;
-      up.push_back(raw[i]);
-    }
-  }
-  
-  bool _raw_to_temp_osds(const pg_pool_t& pool, pg_t pg, vector<int>& raw, vector<int>& temp) const {
-    pg = pool.raw_pg_to_pg(pg);
-    map<pg_t,vector<int> >::const_iterator p = pg_temp.find(pg);
-    if (p != pg_temp.end()) {
-      temp.clear();
-      for (unsigned i=0; i<p->second.size(); i++) {
-	if (!exists(p->second[i]) || is_down(p->second[i]))
-	  continue;
-	temp.push_back(p->second[i]);
-      }
-      return true;
-    }
-    return false;
-  }
+  bool _raw_to_temp_osds(const pg_pool_t& pool, pg_t pg, vector<int>& raw, vector<int>& temp) const;
 
 public:
-  int pg_to_osds(pg_t pg, vector<int>& raw) const {
-    const pg_pool_t *pool = get_pg_pool(pg.pool());
-    if (!pool)
-      return 0;
-    return _pg_to_osds(*pool, pg, raw);
-  }
-
-  int pg_to_acting_osds(pg_t pg, vector<int>& acting) const {         // list of osd addr's
-    const pg_pool_t *pool = get_pg_pool(pg.pool());
-    if (!pool)
-      return 0;
-    vector<int> raw;
-    _pg_to_osds(*pool, pg, raw);
-    if (!_raw_to_temp_osds(*pool, pg, raw, acting))
-      _raw_to_up_osds(pg, raw, acting);
-    return acting.size();
-  }
-
-  void pg_to_raw_up(pg_t pg, vector<int>& up) {
-    const pg_pool_t *pool = get_pg_pool(pg.pool());
-    if (!pool)
-      return;
-    vector<int> raw;
-    _pg_to_osds(*pool, pg, raw);
-    _raw_to_up_osds(pg, raw, up);
-  }
-  
-  void pg_to_up_acting_osds(pg_t pg, vector<int>& up, vector<int>& acting) const {
-    const pg_pool_t *pool = get_pg_pool(pg.pool());
-    if (!pool)
-      return;
-    vector<int> raw;
-    _pg_to_osds(*pool, pg, raw);
-    _raw_to_up_osds(pg, raw, up);
-    if (!_raw_to_temp_osds(*pool, pg, raw, acting))
-      acting = up;
-  }
+  int pg_to_osds(pg_t pg, vector<int>& raw) const;
+  int pg_to_acting_osds(pg_t pg, vector<int>& acting) const;
+  void pg_to_raw_up(pg_t pg, vector<int>& up) const;
+  void pg_to_up_acting_osds(pg_t pg, vector<int>& up, vector<int>& acting) const;
 
   int64_t lookup_pg_pool_name(const char *name) {
     if (name_pool.count(name))
@@ -591,7 +464,7 @@ public:
   }
 
   // pg -> primary osd
-  int get_pg_primary(pg_t pg) {
+  int get_pg_primary(pg_t pg) const {
     vector<int> group;
     int nrep = pg_to_osds(pg, group);
     if (nrep)
@@ -600,14 +473,14 @@ public:
   }
 
   // pg -> acting primary osd
-  int get_pg_acting_primary(pg_t pg) {
+  int get_pg_acting_primary(pg_t pg) const {
     vector<int> group;
     int nrep = pg_to_acting_osds(pg, group);
     if (nrep > 0)
       return group[0];
     return -1;  // we fail!
   }
-  int get_pg_acting_tail(pg_t pg) {
+  int get_pg_acting_tail(pg_t pg) const {
     vector<int> group;
     int nrep = pg_to_acting_osds(pg, group);
     if (nrep > 0)
@@ -617,43 +490,17 @@ public:
 
 
   /* what replica # is a given osd? 0 primary, -1 for none. */
-  static int calc_pg_rank(int osd, vector<int>& acting, int nrep=0) {
-    if (!nrep)
-      nrep = acting.size();
-    for (int i=0; i<nrep; i++) 
-      if (acting[i] == osd)
-	return i;
-    return -1;
-  }
-  static int calc_pg_role(int osd, vector<int>& acting, int nrep=0) {
-    if (!nrep)
-      nrep = acting.size();
-    int rank = calc_pg_rank(osd, acting, nrep);
-    
-    if (rank < 0)
-      return PG_ROLE_STRAY;
-    else if (rank == 0) 
-      return PG_ROLE_HEAD;
-    else if (rank == 1) 
-      return PG_ROLE_ACKER;
-    else
-      return PG_ROLE_MIDDLE;
-  }
-  
-  int get_pg_role(pg_t pg, int osd) const {
-    vector<int> group;
-    int nrep = pg_to_osds(pg, group);
-    return calc_pg_role(osd, group, nrep);
-  }
+  static int calc_pg_rank(int osd, vector<int>& acting, int nrep=0);
+  static int calc_pg_role(int osd, vector<int>& acting, int nrep=0);
   
   /* rank is -1 (stray), 0 (primary), 1,2,3,... (replica) */
-  int get_pg_acting_rank(pg_t pg, int osd) {
+  int get_pg_acting_rank(pg_t pg, int osd) const {
     vector<int> group;
     int nrep = pg_to_acting_osds(pg, group);
     return calc_pg_rank(osd, group, nrep);
   }
   /* role is -1 (stray), 0 (primary), 1 (replica) */
-  int get_pg_acting_role(pg_t pg, int osd) {
+  int get_pg_acting_role(pg_t pg, int osd) const {
     vector<int> group;
     int nrep = pg_to_acting_osds(pg, group);
     return calc_pg_role(osd, group, nrep);
@@ -664,9 +511,9 @@ public:
    * handy helpers to build simple maps...
    */
   void build_simple(CephContext *cct, epoch_t e, uuid_d &fsid,
-		    int num_osd, int pg_bits, int pgp_bits, int lpg_bits);
+		    int num_osd, int pg_bits, int pgp_bits);
   void build_simple_from_conf(CephContext *cct, epoch_t e, uuid_d &fsid,
-			      int pg_bits, int pgp_bits, int lpg_bits);
+			      int pg_bits, int pgp_bits);
   static void build_simple_crush_map(CephContext *cct, CrushWrapper& crush,
 				     map<int, const char*>& poolsets, int num_osd);
   static void build_simple_crush_map_from_conf(CephContext *cct, CrushWrapper& crush,
@@ -681,6 +528,7 @@ public:
   void print_tree(ostream& out) const;
 
   string get_flag_string() const;
+  static string get_flag_string(unsigned flags);
   void dump_json(ostream& out) const;
   void dump(Formatter *f) const;
   static void generate_test_instances(list<OSDMap*>& o);
