@@ -21,20 +21,26 @@
 #include <boost/statechart/state.hpp>
 #include <boost/statechart/state_machine.hpp>
 #include <boost/statechart/transition.hpp>
+#include <boost/statechart/event_base.hpp>
+#include <boost/scoped_ptr.hpp>
+#include <tr1/memory>
 
 // re-include our assert to clobber boost's
 #include "include/assert.h" 
 
 #include "include/types.h"
+#include "include/stringify.h"
 #include "osd_types.h"
 #include "include/buffer.h"
 #include "include/xlist.h"
 #include "include/atomic.h"
 
+#include "OpRequest.h"
 #include "OSDMap.h"
 #include "os/ObjectStore.h"
 #include "msg/Messenger.h"
 #include "messages/MOSDRepScrub.h"
+#include "messages/MOSDPGLog.h"
 
 #include "common/DecayCounter.h"
 
@@ -50,16 +56,14 @@ using namespace __gnu_cxx;
 
 //#define DEBUG_RECOVERY_OIDS   // track set of recovering oids explicitly, to find counting bugs
 
-
 class OSD;
-class OpRequest;
+class OSDService;
 class MOSDOp;
 class MOSDSubOp;
 class MOSDSubOpReply;
 class MOSDPGScan;
 class MOSDPGBackfill;
 class MOSDPGInfo;
-class MOSDPGLog;
 
 
 struct PGRecoveryStats {
@@ -114,8 +118,6 @@ struct PGRecoveryStats {
 
 struct PGPool {
   int id;
-  atomic_t nref;
-  int num_pg;
   string name;
   uint64_t auid;
 
@@ -126,19 +128,14 @@ struct PGPool {
   interval_set<snapid_t> newly_removed_snaps;  // newly removed in the last epoch
 
   PGPool(int i, const char *_name, uint64_t au) :
-    id(i), num_pg(0), auid(au) {
+    id(i), auid(au) {
     if (_name)
       name = _name;
   }
 
-  void get() { nref.inc(); }
-  void put() {
-    if (nref.dec() == 0)
-      delete this;
-  }
+  void update(OSDMapRef map);
 };
 
-
 /** PG - Replica Placement Group
  *
  */
@@ -332,10 +329,10 @@ public:
 
   /*** PG ****/
 protected:
-  OSD *osd;
-  PGPool *pool;
-
+  OSDService *osd;
   OSDMapRef osdmap_ref;
+  PGPool pool;
+
   OSDMapRef get_osdmap() const {
     assert(is_locked());
     assert(osdmap_ref);
@@ -359,7 +356,8 @@ public:
   void lock(bool no_lockdep = false);
   void unlock() {
     //generic_dout(0) << this << " " << info.pgid << " unlock" << dendl;
-    osdmap_ref.reset();
+    assert(!dirty_info);
+    assert(!dirty_log);
     _lock.Unlock();
   }
 
@@ -367,6 +365,9 @@ public:
    * *_with_map_lock_held assume that the map_lock is already held */
   void lock_with_map_lock_held(bool no_lockdep = false);
 
+  // assert we still have lock held, and update our map ref
+  void reassert_lock_with_map_lock_held();
+
   void assert_locked() {
     assert(_lock.is_locked());
   }
@@ -394,62 +395,12 @@ public:
   }
 
 
-  list<OpRequest*> op_queue;  // op queue
+  list<OpRequestRef> op_waiters;
+  list<OpRequestRef> op_queue;  // op queue
 
   bool dirty_info, dirty_log;
 
 public:
-  struct Interval {
-    vector<int> up, acting;
-    epoch_t first, last;
-    bool maybe_went_rw;
-
-    Interval() : first(0), last(0), maybe_went_rw(false) {}
-
-    void encode(bufferlist& bl) const {
-      ENCODE_START(2, 2, bl);
-      ::encode(first, bl);
-      ::encode(last, bl);
-      ::encode(up, bl);
-      ::encode(acting, bl);
-      ::encode(maybe_went_rw, bl);
-      ENCODE_FINISH(bl);
-    }
-    void decode(bufferlist::iterator& bl) {
-      DECODE_START_LEGACY_COMPAT_LEN(2, 2, 2, bl);
-      ::decode(first, bl);
-      ::decode(last, bl);
-      ::decode(up, bl);
-      ::decode(acting, bl);
-      ::decode(maybe_went_rw, bl);
-      DECODE_FINISH(bl);
-    }
-    void dump(Formatter *f) const {
-      f->dump_unsigned("first", first);
-      f->dump_unsigned("last", last);
-      f->dump_int("maybe_went_rw", maybe_went_rw ? 1 : 0);
-      f->open_array_section("up");
-      for (vector<int>::const_iterator p = up.begin(); p != up.end(); ++p)
-	f->dump_int("osd", *p);
-      f->close_section();
-      f->open_array_section("acting");
-      for (vector<int>::const_iterator p = acting.begin(); p != acting.end(); ++p)
-	f->dump_int("osd", *p);
-      f->close_section();
-    }
-    static void generate_test_instances(list<Interval*>& o) {
-      o.push_back(new Interval);
-      o.push_back(new Interval);
-      o.back()->up.push_back(1);
-      o.back()->acting.push_back(2);
-      o.back()->acting.push_back(3);
-      o.back()->first = 4;
-      o.back()->last = 5;
-      o.back()->maybe_went_rw = true;
-    }
-  };
-  WRITE_CLASS_ENCODER(Interval)
-
   // pg state
   pg_info_t        info;
   const coll_t coll;
@@ -459,15 +410,16 @@ public:
   OndiskLog   ondisklog;
   pg_missing_t     missing;
   map<hobject_t, set<int> > missing_loc;
+  set<int> missing_loc_sources;           // superset of missing_loc locations
   
   interval_set<snapid_t> snap_collections;
-  map<epoch_t,Interval> past_intervals;
+  map<epoch_t,pg_interval_t> past_intervals;
 
   interval_set<snapid_t> snap_trimq;
 
   /* You should not use these items without taking their respective queue locks
    * (if they have one) */
-  xlist<PG*>::item recovery_item, scrub_item, scrub_finalize_item, snap_trim_item, remove_item, stat_queue_item;
+  xlist<PG*>::item recovery_item, scrub_item, scrub_finalize_item, snap_trim_item, stat_queue_item;
   int recovery_ops_active;
   bool waiting_on_backfill;
 #ifdef DEBUG_RECOVERY_OIDS
@@ -503,7 +455,7 @@ public:
 
     bool pg_down;   /// some down osds are included in @a cur; the DOWN pg state bit should be set.
     PriorSet(const OSDMap &osdmap,
-	     const map<epoch_t, Interval> &past_intervals,
+	     const map<epoch_t, pg_interval_t> &past_intervals,
 	     const vector<int> &up,
 	     const vector<int> &acting,
 	     const pg_info_t &info,
@@ -522,20 +474,22 @@ public:
   struct RecoveryCtx {
     utime_t start_time;
     map< int, map<pg_t, pg_query_t> > *query_map;
-    map< int, MOSDPGInfo* > *info_map;
-    map< int, vector<pg_info_t> > *notify_list;
-    list< Context* > *context_list;
+    map< int, vector<pair<pg_notify_t, pg_interval_map_t> > > *info_map;
+    map< int, vector<pair<pg_notify_t, pg_interval_map_t> > > *notify_list;
+    C_Contexts *on_applied;
+    C_Contexts *on_safe;
     ObjectStore::Transaction *transaction;
-    RecoveryCtx() : query_map(0), info_map(0), notify_list(0),
-		    context_list(0), transaction(0) {}
     RecoveryCtx(map< int, map<pg_t, pg_query_t> > *query_map,
-		map< int, MOSDPGInfo* > *info_map,
-		map< int, vector<pg_info_t> > *notify_list,
-		list< Context* > *context_list,
+		map< int, vector<pair<pg_notify_t, pg_interval_map_t> > > *info_map,
+		map< int, vector<pair<pg_notify_t, pg_interval_map_t> > > *notify_list,
+		C_Contexts *on_applied,
+		C_Contexts *on_safe,
 		ObjectStore::Transaction *transaction)
       : query_map(query_map), info_map(info_map), 
 	notify_list(notify_list),
-	context_list(context_list), transaction(transaction) {}
+	on_applied(on_applied),
+	on_safe(on_safe),
+	transaction(transaction) {}
   };
 
   struct NamedState {
@@ -558,6 +512,7 @@ protected:
   set<int>    stray_set;   // non-acting osds that have PG data.
   eversion_t  oldest_update; // acting: lowest (valid) last_update in active set
   map<int,pg_info_t>    peer_info;   // info from peers (stray or prior)
+  set<int> peer_purged; // peers purged
   map<int,pg_missing_t> peer_missing;
   set<int>             peer_log_requested;  // logs i've requested (and start stamps)
   set<int>             peer_missing_requested;
@@ -573,9 +528,12 @@ protected:
 
 
   /* heartbeat peers */
+  void set_probe_targets(const set<int> &probe_set);
+  void clear_probe_targets();
 public:
   Mutex heartbeat_peer_lock;
   set<int> heartbeat_peers;
+  set<int> probe_targets;
 
 protected:
   /**
@@ -631,6 +589,22 @@ protected:
       else
 	begin = objects.begin()->first;
     }
+
+    /// dump
+    void dump(Formatter *f) const {
+      f->dump_stream("begin") << begin;
+      f->dump_stream("end") << end;
+      f->open_array_section("objects");
+      for (map<hobject_t, eversion_t>::const_iterator i = objects.begin();
+	   i != objects.end();
+	   ++i) {
+	f->open_object_section("object");
+	f->dump_stream("object") << i->first;
+	f->dump_stream("version") << i->second;
+	f->close_section();
+      }
+      f->close_section();
+    }
   };
   
   BackfillInterval backfill_info;
@@ -648,14 +622,16 @@ protected:
 
 
   // pg waiters
-  list<OpRequest*>            waiting_for_active;
-  list<OpRequest*>            waiting_for_all_missing;
-  map<hobject_t, list<OpRequest*> > waiting_for_missing_object,
+  bool flushed;
+  list<OpRequestRef>            waiting_for_active;
+  list<OpRequestRef>            waiting_for_all_missing;
+  map<hobject_t, list<OpRequestRef> > waiting_for_missing_object,
                                         waiting_for_degraded_object;
-  map<eversion_t,list<OpRequest*> > waiting_for_ondisk;
-  map<eversion_t,OpRequest*>   replay_queue;
+  map<eversion_t,list<OpRequestRef> > waiting_for_ack, waiting_for_ondisk;
+  map<eversion_t,OpRequestRef>   replay_queue;
 
-  void requeue_object_waiters(map<hobject_t, list<OpRequest*> >& m);
+  void requeue_object_waiters(map<hobject_t, list<OpRequestRef> >& m);
+  void requeue_ops(list<OpRequestRef> &l);
 
   // stats
   Mutex pg_stats_lock;
@@ -663,7 +639,7 @@ protected:
   pg_stat_t pg_stats_stable;
 
   // for ordering writes
-  ObjectStore::Sequencer osr;
+  std::tr1::shared_ptr<ObjectStore::Sequencer> osr;
 
   void update_stats();
   void clear_stats();
@@ -685,6 +661,7 @@ public:
   
   bool needs_recovery() const;
 
+  void mark_clean();  ///< mark an active pg clean
   void generate_past_intervals();
   void trim_past_intervals();
   void build_prior(std::auto_ptr<PriorSet> &prior_set);
@@ -695,6 +672,7 @@ public:
 
   bool all_unfound_are_queried_or_lost(const OSDMapRef osdmap) const;
   virtual void mark_all_unfound_lost(int how) = 0;
+  virtual void dump_recovery_info(Formatter *f) const = 0;
 
   bool calc_min_last_complete_ondisk() {
     eversion_t min = last_complete_ondisk;
@@ -717,8 +695,17 @@ public:
 			pg_missing_t& omissing, int from);
   void proc_master_log(ObjectStore::Transaction& t, pg_info_t &oinfo, pg_log_t &olog,
 		       pg_missing_t& omissing, int from);
-  bool proc_replica_info(int from, pg_info_t &info);
+  bool proc_replica_info(int from, const pg_info_t &info);
   bool merge_old_entry(ObjectStore::Transaction& t, pg_log_entry_t& oe);
+
+  /**
+   * Merges authoratative log/info into current log/info/store
+   *
+   * @param [in,out] t used to delete obsolete objects
+   * @param [in,out] oinfo recieved authoritative info
+   * @param [in,out] olog recieved authoritative log
+   * @param [in] from peer which sent the information
+   */
   void merge_log(ObjectStore::Transaction& t, pg_info_t &oinfo, pg_log_t &olog, int from);
   void rewind_divergent_log(ObjectStore::Transaction& t, eversion_t newhead);
   bool search_for_missing(const pg_info_t &oinfo, const pg_missing_t *omissing,
@@ -736,9 +723,11 @@ public:
   bool choose_acting(int& newest_update_osd);
   void build_might_have_unfound();
   void replay_queued_ops();
-  void activate(ObjectStore::Transaction& t, list<Context*>& tfin,
+  void activate(ObjectStore::Transaction& t,
+		epoch_t query_epoch,
+		list<Context*>& tfin,
 		map< int, map<pg_t,pg_query_t> >& query_map,
-		map<int, MOSDPGInfo*> *activator_map=0);
+		map<int, vector<pair<pg_notify_t, pg_interval_map_t> > > *activator_map=0);
   void _activate_committed(epoch_t e, entity_inst_t& primary);
   void all_activated_and_committed();
 
@@ -767,7 +756,7 @@ public:
   void clear_recovery_state();
   virtual void _clear_recovery_state() = 0;
   void defer_recovery();
-  virtual void check_recovery_op_pulls(const OSDMapRef newmap) = 0;
+  virtual void check_recovery_sources(const OSDMapRef newmap) = 0;
   void start_recovery_op(const hobject_t& soid);
   void finish_recovery_op(const hobject_t& soid, bool dequeue=false);
 
@@ -781,9 +770,12 @@ public:
   // -- scrub --
   set<int> scrub_reserved_peers;
   map<int,ScrubMap> scrub_received_maps;
-  bool finalizing_scrub; 
+  bool finalizing_scrub;
+  bool scrub_block_writes;
+  bool scrub_active;
   bool scrub_reserved, scrub_reserve_failed;
   int scrub_waiting_on;
+  set<int> scrub_waiting_on_whom;
   epoch_t scrub_epoch_start;
   ScrubMap primary_scrubmap;
   MOSDRepScrub *active_rep_scrub;
@@ -806,6 +798,8 @@ public:
   void build_scrub_map(ScrubMap &map);
   void build_inc_scrub_map(ScrubMap &map, eversion_t v);
   virtual int _scrub(ScrubMap &map, int& errors, int& fixed) { return 0; }
+  virtual coll_t get_temp_coll() = 0;
+  virtual bool have_temp_coll() = 0;
   void clear_scrub_reserved();
   void scrub_reserve_replicas();
   void scrub_unreserve_replicas();
@@ -813,15 +807,174 @@ public:
   bool sched_scrub();
 
   void replica_scrub(class MOSDRepScrub *op);
-  void sub_op_scrub_map(OpRequest *op);
-  void sub_op_scrub_reserve(OpRequest *op);
-  void sub_op_scrub_reserve_reply(OpRequest *op);
-  void sub_op_scrub_unreserve(OpRequest *op);
-  void sub_op_scrub_stop(OpRequest *op);
+  void sub_op_scrub_map(OpRequestRef op);
+  void sub_op_scrub_reserve(OpRequestRef op);
+  void sub_op_scrub_reserve_reply(OpRequestRef op);
+  void sub_op_scrub_unreserve(OpRequestRef op);
+  void sub_op_scrub_stop(OpRequestRef op);
 
 
   // -- recovery state --
 
+  class CephPeeringEvt {
+    epoch_t epoch_sent;
+    epoch_t epoch_requested;
+    boost::intrusive_ptr< const boost::statechart::event_base > evt;
+    string desc;
+  public:
+    template <class T>
+    CephPeeringEvt(epoch_t epoch_sent,
+		   epoch_t epoch_requested,
+		   const T &evt_) :
+      epoch_sent(epoch_sent), epoch_requested(epoch_requested),
+      evt(evt_.intrusive_from_this()) {
+      stringstream out;
+      out << "epoch_sent: " << epoch_sent
+	  << " epoch_requested: " << epoch_requested << " ";
+      evt_.print(&out);
+      desc = out.str();
+    }
+    epoch_t get_epoch_sent() { return epoch_sent; }
+    epoch_t get_epoch_requested() { return epoch_requested; }
+    const boost::statechart::event_base &get_event() { return *evt; }
+    string get_desc() { return desc; }
+  };
+  typedef std::tr1::shared_ptr<CephPeeringEvt> CephPeeringEvtRef;
+  list<CephPeeringEvtRef> peering_queue;  // op queue
+  list<CephPeeringEvtRef> peering_waiters;
+
+  struct QueryState : boost::statechart::event< QueryState > {
+    Formatter *f;
+    QueryState(Formatter *f) : f(f) {}
+    void print(std::ostream *out) const {
+      *out << "Query" << std::endl;
+    }
+  };
+
+  struct MInfoRec : boost::statechart::event< MInfoRec > {
+    int from;
+    pg_info_t info;
+    epoch_t msg_epoch;
+    MInfoRec(int from, pg_info_t &info, epoch_t msg_epoch) :
+      from(from), info(info), msg_epoch(msg_epoch) {}
+    void print(std::ostream *out) const {
+      *out << "MInfoRec from " << from
+	   << " info: " << info
+	   << std::endl;
+    }
+  };
+
+  struct MLogRec : boost::statechart::event< MLogRec > {
+    int from;
+    boost::intrusive_ptr<MOSDPGLog> msg;
+    MLogRec(int from, MOSDPGLog *msg) :
+      from(from), msg(msg) {}
+    void print(std::ostream *out) const {
+      *out << "MLogRec from " << from
+	   << std::endl;
+    }
+  };
+
+  struct MNotifyRec : boost::statechart::event< MNotifyRec > {
+    int from;
+    pg_notify_t notify;
+    MNotifyRec(int from, pg_notify_t &notify) :
+      from(from), notify(notify) {}
+    void print(std::ostream *out) const {
+      *out << "MNotifyRec from " << from
+	   << " notify: " << notify
+	   << std::endl;
+    }
+  };
+
+  struct MQuery : boost::statechart::event< MQuery > {
+    int from;
+    pg_query_t query;
+    epoch_t query_epoch;
+    MQuery(int from, const pg_query_t &query, epoch_t query_epoch):
+      from(from), query(query), query_epoch(query_epoch) {}
+    void print(std::ostream *out) const {
+      *out << "MQuery from " << from
+	   << " query_epoch " << query_epoch
+	   << " query: " << query
+	   << std::endl;
+    }
+  };
+
+  struct AdvMap : boost::statechart::event< AdvMap > {
+    OSDMapRef osdmap;
+    OSDMapRef lastmap;
+    vector<int> newup, newacting;
+    AdvMap(OSDMapRef osdmap, OSDMapRef lastmap, vector<int>& newup, vector<int>& newacting):
+      osdmap(osdmap), lastmap(lastmap), newup(newup), newacting(newacting) {}
+    void print(std::ostream *out) const {
+      *out << "AdvMap" << std::endl;
+    }
+  };
+
+  struct RecoveryComplete : boost::statechart::event< RecoveryComplete > {
+    RecoveryComplete() : boost::statechart::event< RecoveryComplete >() {}
+    void print(std::ostream *out) const {
+      *out << "RecoveryComplete" << std::endl;
+    }
+  };
+  struct ActMap : boost::statechart::event< ActMap > {
+    ActMap() : boost::statechart::event< ActMap >() {}
+    void print(std::ostream *out) const {
+      *out << "ActMap" << std::endl;
+    }
+  };
+  struct Activate : boost::statechart::event< Activate > {
+    epoch_t query_epoch;
+    Activate(epoch_t q) : boost::statechart::event< Activate >(),
+			  query_epoch(q) {}
+    void print(std::ostream *out) const {
+      *out << "Activate from " << query_epoch << std::endl;
+    }
+  };
+  struct Initialize : boost::statechart::event< Initialize > {
+    Initialize() : boost::statechart::event< Initialize >() {}
+    void print(std::ostream *out) const {
+      *out << "Initialize" << std::endl;
+    }
+  };
+  struct Load : boost::statechart::event< Load > {
+    Load() : boost::statechart::event< Load >() {}
+    void print(std::ostream *out) const {
+      *out << "Load" << std::endl;
+    }
+  };
+  struct GotInfo : boost::statechart::event< GotInfo > {
+    GotInfo() : boost::statechart::event< GotInfo >() {}
+    void print(std::ostream *out) const {
+      *out << "GotInfo" << std::endl;
+    }
+  };
+  struct NeedUpThru : boost::statechart::event< NeedUpThru > {
+    NeedUpThru() : boost::statechart::event< NeedUpThru >() {};
+    void print(std::ostream *out) const {
+      *out << "NeedUpThru" << std::endl;
+    }
+  };
+  struct CheckRepops : boost::statechart::event< CheckRepops > {
+    CheckRepops() : boost::statechart::event< CheckRepops >() {};
+    void print(std::ostream *out) const {
+      *out << "CheckRepops" << std::endl;
+    }
+  };
+  struct NullEvt : boost::statechart::event< NullEvt > {
+    NullEvt() : boost::statechart::event< NullEvt >() {};
+    void print(std::ostream *out) const {
+      *out << "NullEvt" << std::endl;
+    }
+  };
+  struct FlushedEvt : boost::statechart::event< FlushedEvt > {
+    FlushedEvt() : boost::statechart::event< FlushedEvt >() {};
+    void print(std::ostream *out) const {
+      *out << "FlushedEvt" << std::endl;
+    }
+  };
+
   /* Encapsulates PG recovery process */
   class RecoveryState {
     void start_handle(RecoveryCtx *new_ctx) {
@@ -840,71 +993,6 @@ public:
       rctx = 0;
     }
 
-    struct QueryState : boost::statechart::event< QueryState > {
-      Formatter *f;
-      QueryState(Formatter *f) : f(f) {}
-    };
-
-    struct MInfoRec : boost::statechart::event< MInfoRec > {
-      int from;
-      pg_info_t &info;
-      MInfoRec(int from, pg_info_t &info) :
-	from(from), info(info) {}
-    };
-
-    struct MLogRec : boost::statechart::event< MLogRec > {
-      int from;
-      MOSDPGLog *msg;
-      MLogRec(int from, MOSDPGLog *msg) :
-	from(from), msg(msg) {}
-    };
-
-    struct MNotifyRec : boost::statechart::event< MNotifyRec > {
-      int from;
-      pg_info_t &info;
-      MNotifyRec(int from, pg_info_t &info) :
-	from(from), info(info) {}
-    };
-
-    struct MQuery : boost::statechart::event< MQuery > {
-      int from;
-      const pg_query_t &query;
-      epoch_t query_epoch;
-      MQuery(int from, const pg_query_t &query, epoch_t query_epoch):
-	from(from), query(query), query_epoch(query_epoch) {}
-    };
-
-    struct AdvMap : boost::statechart::event< AdvMap > {
-      OSDMapRef osdmap;
-      OSDMapRef lastmap;
-      vector<int> newup, newacting;
-      AdvMap(OSDMapRef osdmap, OSDMapRef lastmap, vector<int>& newup, vector<int>& newacting):
-	osdmap(osdmap), lastmap(lastmap), newup(newup), newacting(newacting) {}
-    };
-
-    struct RecoveryComplete : boost::statechart::event< RecoveryComplete > {
-      RecoveryComplete() : boost::statechart::event< RecoveryComplete >() {}
-    };
-    struct ActMap : boost::statechart::event< ActMap > {
-      ActMap() : boost::statechart::event< ActMap >() {}
-    };
-    struct Activate : boost::statechart::event< Activate > {
-      Activate() : boost::statechart::event< Activate >() {}
-    };
-    struct Initialize : boost::statechart::event< Initialize > {
-      Initialize() : boost::statechart::event< Initialize >() {}
-    };
-    struct Load : boost::statechart::event< Load > {
-      Load() : boost::statechart::event< Load >() {}
-    };
-    struct GotInfo : boost::statechart::event< GotInfo > {
-      GotInfo() : boost::statechart::event< GotInfo >() {}
-    };
-    struct NeedUpThru : boost::statechart::event< NeedUpThru > {
-      NeedUpThru() : boost::statechart::event< NeedUpThru >() {};
-    };
-
-
     /* States */
     struct Initial;
     class RecoveryMachine : public boost::statechart::state_machine< RecoveryMachine, Initial > {
@@ -941,19 +1029,24 @@ public:
 	return state->rctx->query_map;
       }
 
-      map<int, MOSDPGInfo*> *get_info_map() {
+      map<int, vector<pair<pg_notify_t, pg_interval_map_t> > > *get_info_map() {
 	assert(state->rctx->info_map);
 	return state->rctx->info_map;
       }
 
-      list< Context* > *get_context_list() {
-	assert(state->rctx->context_list);
-	return state->rctx->context_list;
+      list< Context* > *get_on_safe_context_list() {
+	assert(state->rctx->on_safe);
+	return &(state->rctx->on_safe->contexts);
       }
 
-      void send_notify(int to, const pg_info_t &info) {
+      list< Context * > *get_on_applied_context_list() {
+	assert(state->rctx->on_applied);
+	return &(state->rctx->on_applied->contexts);
+      }
+
+      void send_notify(int to, const pg_notify_t &info, const pg_interval_map_t &pi) {
 	assert(state->rctx->notify_list);
-	(*state->rctx->notify_list)[to].push_back(info);
+	(*state->rctx->notify_list)[to].push_back(make_pair(info, pi));
       }
     };
     friend class RecoveryMachine;
@@ -979,17 +1072,18 @@ public:
       void exit();
 
       typedef boost::mpl::list <
-	boost::statechart::transition< Initialize, Started >,
+	boost::statechart::transition< Initialize, Reset >,
 	boost::statechart::transition< Load, Reset >,
-	boost::statechart::custom_reaction< MNotifyRec >,
-	boost::statechart::custom_reaction< MInfoRec >,
-	boost::statechart::custom_reaction< MLogRec >,
+	boost::statechart::custom_reaction< NullEvt >,
 	boost::statechart::transition< boost::statechart::event_base, Crashed >
 	> reactions;
 
       boost::statechart::result react(const MNotifyRec&);
       boost::statechart::result react(const MInfoRec&);
       boost::statechart::result react(const MLogRec&);
+      boost::statechart::result react(const boost::statechart::event_base&) {
+	return discard_event();
+      }
     };
 
     struct Reset : boost::statechart::state< Reset, RecoveryMachine >, NamedState {
@@ -1000,11 +1094,17 @@ public:
 	boost::statechart::custom_reaction< QueryState >,
 	boost::statechart::custom_reaction< AdvMap >,
 	boost::statechart::custom_reaction< ActMap >,
+	boost::statechart::custom_reaction< NullEvt >,
+	boost::statechart::custom_reaction< FlushedEvt >,
 	boost::statechart::transition< boost::statechart::event_base, Crashed >
 	> reactions;
       boost::statechart::result react(const QueryState& q);
       boost::statechart::result react(const AdvMap&);
       boost::statechart::result react(const ActMap&);
+      boost::statechart::result react(const FlushedEvt&);
+      boost::statechart::result react(const boost::statechart::event_base&) {
+	return discard_event();
+      }
     };
 
     struct Start;
@@ -1016,10 +1116,16 @@ public:
       typedef boost::mpl::list <
 	boost::statechart::custom_reaction< QueryState >,
 	boost::statechart::custom_reaction< AdvMap >,
+	boost::statechart::custom_reaction< NullEvt >,
+	boost::statechart::custom_reaction< FlushedEvt >,
 	boost::statechart::transition< boost::statechart::event_base, Crashed >
 	> reactions;
       boost::statechart::result react(const QueryState& q);
       boost::statechart::result react(const AdvMap&);
+      boost::statechart::result react(const FlushedEvt&);
+      boost::statechart::result react(const boost::statechart::event_base&) {
+	return discard_event();
+      }
     };
 
     struct MakePrimary : boost::statechart::event< MakePrimary > {
@@ -1094,6 +1200,7 @@ public:
 
     struct Peering : boost::statechart::state< Peering, Primary, GetInfo >, NamedState {
       std::auto_ptr< PriorSet > prior_set;
+      bool flushed;
 
       Peering(my_context ctx);
       void exit();
@@ -1138,13 +1245,15 @@ public:
 	boost::statechart::custom_reaction< ActMap >,
 	boost::statechart::custom_reaction< MQuery >,
 	boost::statechart::custom_reaction< MInfoRec >,
-	boost::statechart::custom_reaction< MLogRec >
+	boost::statechart::custom_reaction< MLogRec >,
+	boost::statechart::custom_reaction< Activate >
 	> reactions;
       boost::statechart::result react(const QueryState& q);
       boost::statechart::result react(const MInfoRec& infoevt);
       boost::statechart::result react(const MLogRec& logevt);
       boost::statechart::result react(const ActMap&);
       boost::statechart::result react(const MQuery&);
+      boost::statechart::result react(const Activate&);
     };
 
     struct Stray : boost::statechart::state< Stray, Started >, NamedState {
@@ -1157,8 +1266,7 @@ public:
 	boost::statechart::custom_reaction< MQuery >,
 	boost::statechart::custom_reaction< MLogRec >,
 	boost::statechart::custom_reaction< MInfoRec >,
-	boost::statechart::custom_reaction< ActMap >,
-	boost::statechart::transition< Activate, ReplicaActive >
+	boost::statechart::custom_reaction< ActMap >
 	> reactions;
       boost::statechart::result react(const MQuery& query);
       boost::statechart::result react(const MLogRec& logevt);
@@ -1191,10 +1299,9 @@ public:
 
     struct GetLog : boost::statechart::state< GetLog, Peering >, NamedState {
       int newest_update_osd;
-      MOSDPGLog *msg;
+      boost::intrusive_ptr<MOSDPGLog> msg;
 
       GetLog(my_context ctx);
-      ~GetLog();
       void exit();
 
       typedef boost::mpl::list <
@@ -1208,6 +1315,7 @@ public:
     };
 
     struct WaitUpThru;
+    struct WaitFlushedPeering;
 
     struct GetMissing : boost::statechart::state< GetMissing, Peering >, NamedState {
       set<int> peer_missing_requested;
@@ -1218,12 +1326,25 @@ public:
       typedef boost::mpl::list <
 	boost::statechart::custom_reaction< QueryState >,
 	boost::statechart::custom_reaction< MLogRec >,
-	boost::statechart::transition< NeedUpThru, WaitUpThru >
+	boost::statechart::transition< NeedUpThru, WaitUpThru >,
+	boost::statechart::transition< CheckRepops, WaitFlushedPeering>
 	> reactions;
       boost::statechart::result react(const QueryState& q);
       boost::statechart::result react(const MLogRec& logevt);
     };
 
+    struct WaitFlushedPeering :
+      boost::statechart::state< WaitFlushedPeering, Peering>, NamedState {
+      WaitFlushedPeering(my_context ctx);
+      void exit() {}
+      typedef boost::mpl::list <
+	boost::statechart::custom_reaction< QueryState >,
+	boost::statechart::custom_reaction< FlushedEvt >
+      > reactions;
+      boost::statechart::result react(const FlushedEvt& evt);
+      boost::statechart::result react(const QueryState& q);
+    };
+
     struct WaitUpThru : boost::statechart::state< WaitUpThru, Peering >, NamedState {
       WaitUpThru(my_context ctx);
       void exit();
@@ -1231,6 +1352,7 @@ public:
       typedef boost::mpl::list <
 	boost::statechart::custom_reaction< QueryState >,
 	boost::statechart::custom_reaction< ActMap >,
+	boost::statechart::transition< CheckRepops, WaitFlushedPeering>,
 	boost::statechart::custom_reaction< MLogRec >
 	> reactions;
       boost::statechart::result react(const QueryState& q);
@@ -1248,56 +1370,28 @@ public:
       machine.initiate();
     }
 
-    void handle_notify(int from, pg_info_t& i, RecoveryCtx *ctx);
-    void handle_info(int from, pg_info_t& i, RecoveryCtx *ctx);
-    void handle_log(int from,
-		    MOSDPGLog *msg,
-		    RecoveryCtx *ctx);
-    void handle_query(int from, const pg_query_t& q,
-		      epoch_t query_epoch,
-		      RecoveryCtx *ctx);
-    void handle_advance_map(OSDMapRef osdmap, OSDMapRef lastmap,
-			    vector<int>& newup, vector<int>& newacting, 
-			    RecoveryCtx *ctx);
-    void handle_activate_map(RecoveryCtx *ctx);
-    void handle_recovery_complete(RecoveryCtx *ctx);
-    void handle_create(RecoveryCtx *ctx);
-    void handle_loaded(RecoveryCtx *ctx);
-    void handle_query_state(Formatter *f);
+    void handle_event(const boost::statechart::event_base &evt,
+		      RecoveryCtx *rctx) {
+      start_handle(rctx);
+      machine.process_event(evt);
+      end_handle();
+    }
+
+    void handle_event(CephPeeringEvtRef evt,
+		      RecoveryCtx *rctx) {
+      start_handle(rctx);
+      machine.process_event(evt->get_event());
+      end_handle();
+    }
+
   } recovery_state;
 
 
- public:  
-  PG(OSD *o, PGPool *_pool, pg_t p, const hobject_t& loid, const hobject_t& ioid) : 
-    osd(o), pool(_pool),
-    _lock("PG::_lock"),
-    ref(0), deleting(false), dirty_info(false), dirty_log(false),
-    info(p), coll(p), log_oid(loid), biginfo_oid(ioid),
-    recovery_item(this), scrub_item(this), scrub_finalize_item(this), snap_trim_item(this), remove_item(this), stat_queue_item(this),
-    recovery_ops_active(0),
-    waiting_on_backfill(0),
-    role(0),
-    state(0),
-    need_up_thru(false),
-    need_flush(false),
-    last_peering_reset(0),
-    heartbeat_peer_lock("PG::heartbeat_peer_lock"),
-    backfill_target(-1),
-    pg_stats_lock("PG::pg_stats_lock"),
-    pg_stats_valid(false),
-    finish_sync_event(NULL),
-    finalizing_scrub(false),
-    scrub_reserved(false), scrub_reserve_failed(false),
-    scrub_waiting_on(0),
-    active_rep_scrub(0),
-    recovery_state(this)
-  {
-    pool->get();
-  }
-  virtual ~PG() {
-    pool->put();
-  }
-  
+ public:
+  PG(OSDService *o, OSDMapRef curmap,
+     PGPool pool, pg_t p, const hobject_t& loid, const hobject_t& ioid);
+  virtual ~PG();
+
  private:
   // Prevent copying
   PG(const PG& rhs);
@@ -1312,7 +1406,7 @@ public:
   int        get_role() const { return role; }
   void       set_role(int r) { role = r; }
 
-  bool       is_primary() const { return role == PG_ROLE_HEAD; }
+  bool       is_primary() const { return role == 0; }
   bool       is_replica() const { return role > 0; }
 
   epoch_t get_last_peering_reset() const { return last_peering_reset; }
@@ -1338,7 +1432,7 @@ public:
   bool  is_empty() const { return info.last_update == eversion_t(0,0); }
 
   void init(int role, vector<int>& up, vector<int>& acting, pg_history_t& history,
-	    ObjectStore::Transaction *t);
+	    pg_interval_map_t& pim, ObjectStore::Transaction *t);
 
   // pg on-disk state
   void do_pending_flush();
@@ -1346,6 +1440,8 @@ public:
   void write_info(ObjectStore::Transaction& t);
   void write_log(ObjectStore::Transaction& t);
 
+  void write_if_dirty(ObjectStore::Transaction& t);
+
   void add_log_entry(pg_log_entry_t& e, bufferlist& log_bl);
   void append_log(vector<pg_log_entry_t>& logv, eversion_t trim_to, ObjectStore::Transaction &t);
 
@@ -1376,61 +1472,70 @@ public:
   void start_peering_interval(const OSDMapRef lastmap,
 			      const vector<int>& newup,
 			      const vector<int>& newacting);
+  void start_flush(ObjectStore::Transaction *t,
+		   list<Context *> *on_applied,
+		   list<Context *> *on_safe);
   void set_last_peering_reset();
 
+  void update_history_from_master(pg_history_t new_history);
   void fulfill_info(int from, const pg_query_t &query, 
 		    pair<int, pg_info_t> &notify_info);
   void fulfill_log(int from, const pg_query_t &query, epoch_t query_epoch);
   bool acting_up_affected(const vector<int>& newup, const vector<int>& newacting);
+
+  // OpRequest queueing
+  bool can_discard_op(OpRequestRef op);
+  bool can_discard_scan(OpRequestRef op);
+  bool can_discard_subop(OpRequestRef op);
+  bool can_discard_backfill(OpRequestRef op);
+  bool can_discard_request(OpRequestRef op);
+
+  bool must_delay_request(OpRequestRef op);
+  void queue_op(OpRequestRef op);
+
   bool old_peering_msg(epoch_t reply_epoch, epoch_t query_epoch);
+  bool old_peering_evt(CephPeeringEvtRef evt) {
+    return old_peering_msg(evt->get_epoch_sent(), evt->get_epoch_requested());
+  }
+  bool require_same_or_newer_map(epoch_t e) {
+    return e <= get_osdmap()->get_epoch();
+  }
 
   // recovery bits
-  void handle_notify(int from, pg_info_t& i, RecoveryCtx *rctx) {
-    recovery_state.handle_notify(from, i, rctx);
-  }
-  void handle_info(int from, pg_info_t& i, RecoveryCtx *rctx) {
-    recovery_state.handle_info(from, i, rctx);
-  }
-  void handle_log(int from,
-		  MOSDPGLog *msg,
-		  RecoveryCtx *rctx) {
-    recovery_state.handle_log(from, msg, rctx);
-  }
-  void handle_query(int from, const pg_query_t& q,
-		    epoch_t query_epoch,
-		    RecoveryCtx *rctx) {
-    recovery_state.handle_query(from, q, query_epoch, rctx);
-  }
+  void take_waiters();
+  void queue_peering_event(CephPeeringEvtRef evt);
+  void handle_peering_event(CephPeeringEvtRef evt, RecoveryCtx *rctx);
+  void queue_notify(epoch_t msg_epoch, epoch_t query_epoch,
+		    int from, pg_notify_t& i);
+  void queue_info(epoch_t msg_epoch, epoch_t query_epoch,
+		  int from, pg_info_t& i);
+  void queue_log(epoch_t msg_epoch, epoch_t query_epoch, int from,
+		 MOSDPGLog *msg);
+  void queue_query(epoch_t msg_epoch, epoch_t query_epoch,
+		   int from, const pg_query_t& q);
+  void queue_null(epoch_t msg_epoch, epoch_t query_epoch);
+  void queue_flushed(epoch_t started_at);
   void handle_advance_map(OSDMapRef osdmap, OSDMapRef lastmap,
 			  vector<int>& newup, vector<int>& newacting,
-			  RecoveryCtx *rctx) {
-    recovery_state.handle_advance_map(osdmap, lastmap, newup, newacting, rctx);
-  }
-  void handle_activate_map(RecoveryCtx *rctx) {
-    recovery_state.handle_activate_map(rctx);
-  }
-  void handle_recovery_complete(RecoveryCtx *rctx) {
-    recovery_state.handle_recovery_complete(rctx);
-  }
-  void handle_create(RecoveryCtx *rctx) {
-    recovery_state.handle_create(rctx);
-  }
-  void handle_loaded(RecoveryCtx *rctx) {
-    recovery_state.handle_loaded(rctx);
-  }
+			  RecoveryCtx *rctx);
+  void handle_activate_map(RecoveryCtx *rctx);
+  void handle_recovery_complete(RecoveryCtx *rctx);
+  void handle_create(RecoveryCtx *rctx);
+  void handle_loaded(RecoveryCtx *rctx);
+  void handle_query_state(Formatter *f);
 
-  void on_removal();
+  virtual void on_removal() = 0;
 
 
   // abstract bits
-  void do_request(OpRequest *op);
+  void do_request(OpRequestRef op);
 
-  virtual void do_op(OpRequest *op) = 0;
-  virtual void do_sub_op(OpRequest *op) = 0;
-  virtual void do_sub_op_reply(OpRequest *op) = 0;
-  virtual void do_scan(OpRequest *op) = 0;
-  virtual void do_backfill(OpRequest *op) = 0;
-  virtual bool snap_trimmer() = 0;
+  virtual void do_op(OpRequestRef op) = 0;
+  virtual void do_sub_op(OpRequestRef op) = 0;
+  virtual void do_sub_op_reply(OpRequestRef op) = 0;
+  virtual void do_scan(OpRequestRef op) = 0;
+  virtual void do_backfill(OpRequestRef op) = 0;
+  virtual void snap_trimmer() = 0;
 
   virtual int do_command(vector<string>& cmd, ostream& ss,
 			 bufferlist& idata, bufferlist& odata) = 0;
@@ -1455,18 +1560,13 @@ public:
 				    utime_t expire) = 0;
 };
 
-WRITE_CLASS_ENCODER(PG::Interval)
 WRITE_CLASS_ENCODER(PG::OndiskLog)
 
-inline ostream& operator<<(ostream& out, const PG::Interval& i)
-{
-  out << "interval(" << i.first << "-" << i.last << " " << i.up << "/" << i.acting;
-  if (i.maybe_went_rw)
-    out << " maybe_went_rw";
-  out << ")";
-  return out;
-}
-
 ostream& operator<<(ostream& out, const PG& pg);
 
+namespace boost {
+  void intrusive_ptr_add_ref(PG *pg);
+  void intrusive_ptr_release(PG *pg);
+};
+
 #endif
