@@ -17,7 +17,7 @@
 
 #include <sstream>
 #include <stdio.h>
-#include <stdexcept>
+#include <memory>
 
 #include "msg/msg_types.h"
 #include "include/types.h"
@@ -36,6 +36,8 @@
 #define CEPH_OSD_FEATURE_INCOMPAT_OLOC CompatSet::Feature(3, "object locator")
 #define CEPH_OSD_FEATURE_INCOMPAT_LEC  CompatSet::Feature(4, "last_epoch_clean")
 #define CEPH_OSD_FEATURE_INCOMPAT_CATEGORIES  CompatSet::Feature(5, "categories")
+#define CEPH_OSD_FEATURE_INCOMPAT_HOBJECTPOOL  CompatSet::Feature(6, "hobjectpool")
+#define CEPH_OSD_FEATURE_INCOMPAT_BIGINFO CompatSet::Feature(7, "biginfo")
 
 
 typedef hobject_t collection_list_handle_t;
@@ -100,28 +102,21 @@ namespace __gnu_cxx {
 // does it go in.
 struct object_locator_t {
   int64_t pool;
-  int32_t preferred;
   string key;
 
   explicit object_locator_t()
-    : pool(-1), preferred(-1) {}
+    : pool(-1) {}
   explicit object_locator_t(int64_t po)
-    : pool(po), preferred(-1) {}
-  explicit object_locator_t(int64_t po, int pre)
-    : pool(po), preferred(pre) {}
-  explicit object_locator_t(int64_t po, int pre, string s)
-    : pool(po), preferred(pre), key(s) {}
+    : pool(po) {}
+  explicit object_locator_t(int64_t po, string s)
+    : pool(po), key(s) {}
 
   int get_pool() const {
     return pool;
   }
-  int get_preferred() const {
-    return preferred;
-  }
 
   void clear() {
     pool = -1;
-    preferred = -1;
     key = "";
   }
 
@@ -133,7 +128,7 @@ struct object_locator_t {
 WRITE_CLASS_ENCODER(object_locator_t)
 
 inline bool operator==(const object_locator_t& l, const object_locator_t& r) {
-  return l.pool == r.pool && l.preferred == r.preferred && l.key == r.key;
+  return l.pool == r.pool && l.key == r.key;
 }
 inline bool operator!=(const object_locator_t& l, const object_locator_t& r) {
   return !(l == r);
@@ -142,8 +137,6 @@ inline bool operator!=(const object_locator_t& l, const object_locator_t& r) {
 inline ostream& operator<<(ostream& out, const object_locator_t& loc)
 {
   out << "@" << loc.pool;
-  if (loc.preferred >= 0)
-    out << "p" << loc.preferred;
   if (loc.key.length())
     out << ":" << loc.key;
   return out;
@@ -314,7 +307,6 @@ namespace __gnu_cxx {
 class coll_t {
 public:
   const static coll_t META_COLL;
-  const static coll_t TEMP_COLL;
 
   coll_t()
     : str("meta")
@@ -328,6 +320,14 @@ public:
     : str(pg_and_snap_to_str(pgid, snap))
   { }
 
+  static coll_t make_temp_coll(pg_t pgid) {
+    return coll_t(pg_to_tmp_str(pgid));
+  }
+
+  static coll_t make_removal_coll(uint64_t seq, pg_t pgid) {
+    return coll_t(seq_to_removal_str(seq, pgid));
+  }
+
   const std::string& to_str() const {
     return str;
   }
@@ -341,6 +341,8 @@ public:
   }
 
   bool is_pg(pg_t& pgid, snapid_t& snap) const;
+  bool is_temp(pg_t& pgid) const;
+  bool is_removal(uint64_t *seq, pg_t *pgid) const;
   void encode(bufferlist& bl) const;
   void decode(bufferlist::iterator& bl);
   inline bool operator==(const coll_t& rhs) const {
@@ -359,6 +361,16 @@ private:
     oss << p << "_" << s;
     return oss.str();
   }
+  static std::string pg_to_tmp_str(pg_t p) {
+    std::ostringstream oss;
+    oss << p << "_TEMP";
+    return oss.str();
+  }
+  static std::string seq_to_removal_str(uint64_t seq, pg_t pgid) {
+    std::ostringstream oss;
+    oss << "FORREMOVAL_" << seq << "_" << pgid;
+    return oss.str();
+  }
 
   std::string str;
 };
@@ -416,6 +428,13 @@ public:
 
   eversion_t(bufferlist& bl) : __pad(0) { decode(bl); }
 
+  static eversion_t max() {
+    eversion_t max;
+    max.version -= 1;
+    max.epoch -= 1;
+    return max;
+  }
+
   operator ceph_eversion() {
     ceph_eversion c;
     c.epoch = epoch;
@@ -596,8 +615,9 @@ struct pg_pool_t {
   __u8 size;                /// number of osds in each pg
   __u8 crush_ruleset;       /// crush placement rule set
   __u8 object_hash;         /// hash mapping object name to ps
+private:
   __u32 pg_num, pgp_num;    /// number of pgs
-  __u32 lpg_num, lpgp_num;  /// number of localized pgs
+public:
   epoch_t last_change;      /// most recent epoch changed, exclusing snapshot changes
   snapid_t snap_seq;        /// seq for per-pool snapshot
   epoch_t snap_epoch;       /// osdmap epoch of last snap
@@ -618,16 +638,16 @@ struct pg_pool_t {
    */
   interval_set<snapid_t> removed_snaps;
 
-  int pg_num_mask, pgp_num_mask, lpg_num_mask, lpgp_num_mask;
+  int pg_num_mask, pgp_num_mask;
 
   pg_pool_t()
     : flags(0), type(0), size(0), crush_ruleset(0), object_hash(0),
-      pg_num(0), pgp_num(0), lpg_num(0), lpgp_num(0),
+      pg_num(0), pgp_num(0),
       last_change(0),
       snap_seq(0), snap_epoch(0),
       auid(0),
       crash_replay_interval(0),
-      pg_num_mask(0), pgp_num_mask(0), lpg_num_mask(0), lpgp_num_mask(0) { }
+      pg_num_mask(0), pgp_num_mask(0) { }
 
   void dump(Formatter *f) const;
 
@@ -653,13 +673,18 @@ struct pg_pool_t {
 
   unsigned get_pg_num() const { return pg_num; }
   unsigned get_pgp_num() const { return pgp_num; }
-  unsigned get_lpg_num() const { return lpg_num; }
-  unsigned get_lpgp_num() const { return lpgp_num; }
 
   unsigned get_pg_num_mask() const { return pg_num_mask; }
   unsigned get_pgp_num_mask() const { return pgp_num_mask; }
-  unsigned get_lpg_num_mask() const { return lpg_num_mask; }
-  unsigned get_lpgp_num_mask() const { return lpgp_num_mask; }
+
+  void set_pg_num(int p) {
+    pg_num = p;
+    calc_pg_masks();
+  }
+  void set_pgp_num(int p) {
+    pgp_num = p;
+    calc_pg_masks();
+  }
 
   static int calc_bits_of(int t);
   void calc_pg_masks();
@@ -672,8 +697,10 @@ struct pg_pool_t {
    *    - removal governed by removed_snaps
    *
    * we know which mode we're using based on whether removed_snaps is empty.
+   * If nothing has been created, both functions report false.
    */
   bool is_pool_snaps_mode() const;
+  bool is_unmanaged_snaps_mode() const;
   bool is_removed_snap(snapid_t s) const;
 
   /*
@@ -931,20 +958,34 @@ struct pg_history_t {
       last_epoch_started(0), last_epoch_clean(0), last_epoch_split(0),
       same_up_since(0), same_interval_since(0), same_primary_since(0) {}
   
-  void merge(const pg_history_t &other) {
+  bool merge(const pg_history_t &other) {
     // Here, we only update the fields which cannot be calculated from the OSDmap.
-    if (epoch_created < other.epoch_created)
+    bool modified = false;
+    if (epoch_created < other.epoch_created) {
       epoch_created = other.epoch_created;
-    if (last_epoch_started < other.last_epoch_started)
+      modified = true;
+    }
+    if (last_epoch_started < other.last_epoch_started) {
       last_epoch_started = other.last_epoch_started;
-    if (last_epoch_clean < other.last_epoch_clean)
+      modified = true;
+    }
+    if (last_epoch_clean < other.last_epoch_clean) {
       last_epoch_clean = other.last_epoch_clean;
-    if (last_epoch_split < other.last_epoch_started)
-      last_epoch_split = other.last_epoch_started;
-    if (other.last_scrub > last_scrub)
+      modified = true;
+    }
+    if (last_epoch_split < other.last_epoch_started) {
+      last_epoch_split = other.last_epoch_started; 
+      modified = true;
+    }
+    if (other.last_scrub > last_scrub) {
       last_scrub = other.last_scrub;
-    if (other.last_scrub_stamp > last_scrub_stamp)
+      modified = true;
+    }
+    if (other.last_scrub_stamp > last_scrub_stamp) {
       last_scrub_stamp = other.last_scrub_stamp;
+      modified = true;
+    }
+    return modified;
   }
 
   void encode(bufferlist& bl) const;
@@ -1027,6 +1068,65 @@ inline ostream& operator<<(ostream& out, const pg_info_t& pgi)
   return out;
 }
 
+struct pg_notify_t {
+  epoch_t query_epoch;
+  epoch_t epoch_sent;
+  pg_info_t info;
+  pg_notify_t() : query_epoch(0), epoch_sent(0) {}
+  pg_notify_t(epoch_t query_epoch,
+	      epoch_t epoch_sent,
+	      const pg_info_t &info)
+    : query_epoch(query_epoch),
+      epoch_sent(epoch_sent),
+      info(info) {}
+  void encode(bufferlist &bl) const;
+  void decode(bufferlist::iterator &p);
+  void dump(Formatter *f) const;
+  static void generate_test_instances(list<pg_notify_t*> &o);
+};
+WRITE_CLASS_ENCODER(pg_notify_t)
+ostream &operator<<(ostream &lhs, const pg_notify_t notify);
+
+
+/**
+ * pg_interval_t - information about a past interval
+ */
+class OSDMap;
+struct pg_interval_t {
+  vector<int> up, acting;
+  epoch_t first, last;
+  bool maybe_went_rw;
+
+  pg_interval_t() : first(0), last(0), maybe_went_rw(false) {}
+
+  void encode(bufferlist& bl) const;
+  void decode(bufferlist::iterator& bl);
+  void dump(Formatter *f) const;
+  static void generate_test_instances(list<pg_interval_t*>& o);
+
+  /**
+   * Integrates a new map into *past_intervals, returns true
+   * if an interval was closed out.
+   */
+  static bool check_new_interval(
+    const vector<int> &old_acting,              ///< [in] acting as of lastmap
+    const vector<int> &new_acting,              ///< [in] acting as of osdmap
+    const vector<int> &old_up,                  ///< [in] up as of lastmap
+    const vector<int> &new_up,                  ///< [in] up as of osdmap
+    epoch_t same_interval_since,                ///< [in] as of osdmap
+    epoch_t last_epoch_clean,                   ///< [in] current
+    std::tr1::shared_ptr<const OSDMap> osdmap,  ///< [in] current map
+    std::tr1::shared_ptr<const OSDMap> lastmap, ///< [in] last map
+    map<epoch_t, pg_interval_t> *past_intervals,///< [out] intervals
+    ostream *out = 0                            ///< [out] debug ostream
+    );
+};
+WRITE_CLASS_ENCODER(pg_interval_t)
+
+ostream& operator<<(ostream& out, const pg_interval_t& i);
+
+typedef map<epoch_t, pg_interval_t> pg_interval_map_t;
+
 
 /** 
  * pg_query_t - used to ask a peer for information about a pg.
@@ -1053,31 +1153,29 @@ struct pg_query_t {
   __s32 type;
   eversion_t since;
   pg_history_t history;
+  epoch_t epoch_sent;
 
   pg_query_t() : type(-1) {}
-  pg_query_t(int t, const pg_history_t& h)
-    : type(t), history(h) {
+  pg_query_t(int t, const pg_history_t& h,
+	     epoch_t epoch_sent)
+    : type(t), history(h),
+      epoch_sent(epoch_sent) {
     assert(t != LOG);
   }
-  pg_query_t(int t, eversion_t s, const pg_history_t& h)
-    : type(t), since(s), history(h) {
+  pg_query_t(int t, eversion_t s, const pg_history_t& h,
+	     epoch_t epoch_sent)
+    : type(t), since(s), history(h),
+      epoch_sent(epoch_sent) {
     assert(t == LOG);
   }
   
-  void encode(bufferlist &bl) const {
-    ::encode(type, bl);
-    ::encode(since, bl);
-    history.encode(bl);
-  }
-  void decode(bufferlist::iterator &bl) {
-    ::decode(type, bl);
-    ::decode(since, bl);
-    history.decode(bl);
-  }
+  void encode(bufferlist &bl, uint64_t features) const;
+  void decode(bufferlist::iterator &bl);
+
   void dump(Formatter *f) const;
   static void generate_test_instances(list<pg_query_t*>& o);
 };
-WRITE_CLASS_ENCODER(pg_query_t)
+WRITE_CLASS_ENCODER_FEATURES(pg_query_t)
 
 inline ostream& operator<<(ostream& out, const pg_query_t& q) {
   out << "query(" << q.get_type_name() << " " << q.since;
@@ -1133,6 +1231,7 @@ struct pg_log_entry_t {
   utime_t     mtime;  // this is the _user_ mtime, mind you
   bufferlist snaps;   // only for clone entries
   bool invalid_hash; // only when decoding sobject_t based entries
+  bool invalid_pool; // only when decoding pool-less hobject based entries
 
   uint64_t offset;   // [soft state] my offset on disk
       
@@ -1143,7 +1242,8 @@ struct pg_log_entry_t {
 		 const osd_reqid_t& rid, const utime_t& mt)
     : op(_op), soid(_soid), version(v),
       prior_version(pv),
-      reqid(rid), mtime(mt), invalid_hash(false), offset(0) {}
+      reqid(rid), mtime(mt), invalid_hash(false), invalid_pool(false),
+      offset(0) {}
       
   bool is_clone() const { return op == CLONE; }
   bool is_modify() const { return op == MODIFY; }
@@ -1258,7 +1358,7 @@ struct pg_log_t {
   ostream& print(ostream& out) const;
 
   void encode(bufferlist &bl) const;
-  void decode(bufferlist::iterator &bl);
+  void decode(bufferlist::iterator &bl, int64_t pool = -1);
   void dump(Formatter *f) const;
   static void generate_test_instances(list<pg_log_t*>& o);
 };
@@ -1316,6 +1416,7 @@ struct pg_missing_t {
   eversion_t have_old(const hobject_t& oid) const;
   void add_next_event(const pg_log_entry_t& e);
   void revise_need(hobject_t oid, eversion_t need);
+  void revise_have(hobject_t oid, eversion_t have);
   void add(const hobject_t& oid, eversion_t need, eversion_t have);
   void rm(const hobject_t& oid, eversion_t v);
   void rm(const std::map<hobject_t, pg_missing_t::item>::iterator &m);
@@ -1328,7 +1429,7 @@ struct pg_missing_t {
   }
 
   void encode(bufferlist &bl) const;
-  void decode(bufferlist::iterator &bl);
+  void decode(bufferlist::iterator &bl, int64_t pool = -1);
   void dump(Formatter *f) const;
   static void generate_test_instances(list<pg_missing_t*>& o);
 };
@@ -1373,7 +1474,7 @@ struct pg_ls_response_t {
   static void generate_test_instances(list<pg_ls_response_t*>& o) {
     o.push_back(new pg_ls_response_t);
     o.push_back(new pg_ls_response_t);
-    o.back()->handle = hobject_t(object_t("hi"), "key", 1, 2);
+    o.back()->handle = hobject_t(object_t("hi"), "key", 1, 2, -1);
     o.back()->entries.push_back(make_pair(object_t("one"), string()));
     o.back()->entries.push_back(make_pair(object_t("two"), string("twokey")));
   }
@@ -1632,7 +1733,7 @@ struct ObjectRecoveryInfo {
 
   static void generate_test_instances(list<ObjectRecoveryInfo*>& o);
   void encode(bufferlist &bl) const;
-  void decode(bufferlist::iterator &bl);
+  void decode(bufferlist::iterator &bl, int64_t pool = -1);
   ostream &print(ostream &out) const;
   void dump(Formatter *f) const;
 };
@@ -1695,7 +1796,7 @@ struct ScrubMap {
   void merge_incr(const ScrubMap &l);
 
   void encode(bufferlist& bl) const;
-  void decode(bufferlist::iterator& bl);
+  void decode(bufferlist::iterator& bl, int64_t pool=-1);
   void dump(Formatter *f) const;
   static void generate_test_instances(list<ScrubMap*>& o);
 };
