@@ -62,46 +62,54 @@ static string dir_info_oid = RBD_INFO;
 
 void usage()
 {
-  cout << "usage: rbd [-n <auth user>] [OPTIONS] <cmd> ...\n"
-       << "where 'pool' is a rados pool name (default is 'rbd') and 'cmd' is one of:\n"
-       << "  <ls | list> [pool-name]                   list rbd images\n"
-       << "  info <--snap=name> [image-name]           show information about image size,\n"
-       << "                                            striping, etc.\n"
-       << "  create <--order=bits> [--size MB] [name]  create an empty image\n"
-       << "  resize [--size MB] [image-name]           resize (expand or contract) image\n"
-       << "  rm [image-name]                           delete an image\n"
-       << "  export <--snap=name> [image-name] [path]  export image to file\n"
-       << "  import [path] [dst-image]                 import image from file (dest defaults\n"
-       << "                                            as the filename part of file)\n"
-       << "  <cp | copy> <--snap=name> [src] [dest]    copy src image to dest\n"
-       << "  <mv | rename> [src] [dest]                rename src image to dest\n"
-       << "  snap ls [image-name]                      dump list of image snapshots\n"
-       << "  snap create <--snap=name> [image-name]    create a snapshot\n"
-       << "  snap rollback <--snap=name> [image-name]  rollback image head to snapshot\n"
-       << "  snap rm <--snap=name> [image-name]        deletes a snapshot\n"
-       << "  snap purge [image-name]                   deletes all snapshots\n"
-       << "  watch [image-name]                        watch events on image\n"
-       << "  map [image-name]                          map the image to a block device\n"
-       << "                                            using the kernel\n"
-       << "  unmap [device]                            unmap a rbd device that was\n"
-       << "                                            mapped by the kernel\n"
-       << "  showmapped                                show the rbd images mapped\n"
-       << "                                            by the kernel\n"
-       << "\n"
-       << "Other input options:\n"
-       << "  -p, --pool <pool>            source pool name\n"
-       << "  --image <image-name>         image name\n"
-       << "  --dest <name>                destination [pool and] image name\n"
-       << "  --snap <snapname>            specify snapshot name\n"
-       << "  --dest-pool <name>           destination pool name\n"
-       << "  --path <path-name>           path name for import/export (if not specified)\n"
-       << "  --size <size in MB>          size parameter for create and resize commands\n"
-       << "  --order <bits>               the object size in bits, such that the objects\n"
-       << "                               are (1 << order) bytes. Default is 22 (4 MB).\n"
-       << "\n"
-       << "For the map command:\n"
-       << "  --user <username>            rados user to authenticate as\n"
-       << "  --secret <path>              file containing secret key for use with cephx\n";
+  cout << 
+"usage: rbd [-n <auth user>] [OPTIONS] <cmd> ...\n"
+"where 'pool' is a rados pool name (default is 'rbd') and 'cmd' is one of:\n"
+"  (ls | list) [pool-name]                     list rbd images\n"
+"  info <image-name>                           show information about image size,\n"
+"                                              striping, etc.\n"
+"  create [--order <bits>] --size <MB> <name>  create an empty image\n"
+"  clone [--order <bits>] <parentsnap> <clonename>\n"
+"                                              clone a snapshot into a COW\n"
+"                                               child image\n"
+"  resize --size <MB> <image-name>             resize (expand or contract) image\n"
+"  rm <image-name>                             delete an image\n"
+"  export <image-name> <path>                  export image to file\n"
+"  import <path> <image-name>                  import image from file\n"
+"                                              (dest defaults)\n"
+"                                              as the filename part of file)\n"
+"  (cp | copy) <src> <dest>                    copy src image to dest\n"
+"  (mv | rename) <src> <dest>                  rename src image to dest\n"
+"  snap ls <image-name>                        dump list of image snapshots\n"
+"  snap create <snap-name>                     create a snapshot\n"
+"  snap rollback <snap-name>                   rollback image to snapshot\n"
+"  snap rm <snap-name>                         deletes a snapshot\n"
+"  snap purge <image-name>                     deletes all snapshots\n"
+"  watch <image-name>                          watch events on image\n"
+"  map <image-name>                            map image to a block device\n"
+"                                              using the kernel\n"
+"  unmap <device>                              unmap a rbd device that was\n"
+"                                              mapped by the kernel\n"
+"  showmapped                                  show the rbd images mapped\n"
+"                                              by the kernel\n"
+"\n"
+"<image-name>, <snap-name> are [pool/]name[@snap], or you may specify\n"
+"individual pieces of names with -p/--pool, --image, and/or --snap.\n"
+"\n"
+"Other input options:\n"
+"  -p, --pool <pool>            source pool name\n"
+"  --image <image-name>         image name\n"
+"  --dest <image-name>          destination [pool and] image name\n"
+"  --snap <snap-name>           snapshot name\n"
+"  --dest-pool <name>           destination pool name\n"
+"  --path <path-name>           path name for import/export\n"
+"  --size <size in MB>          size of image for create and resize\n"
+"  --order <bits>               the object size in bits; object size will be\n"
+"                               (1 << order) bytes. Default is 22 (4 MB).\n"
+"\n"
+"For the map command:\n"
+"  --user <username>            rados user to authenticate as\n"
+"  --secret <path>              file containing secret key for use with cephx\n";
 }
 
 void usage_exit()
@@ -111,20 +119,13 @@ void usage_exit()
   exit(1);
 }
 
-static void print_info(const char *imgname, librbd::image_info_t& info)
+static string feature_str(uint64_t features)
 {
-  cout << "rbd image '" << imgname << "':\n"
-       << "\tsize " << prettybyte_t(info.size) << " in "
-       << info.num_objs << " objects"
-       << std::endl
-       << "\torder " << info.order
-       << " (" << prettybyte_t(info.obj_size) << " objects)"
-       << std::endl
-       << "\tblock_name_prefix: " << info.block_name_prefix
-       << std::endl
-       << "\tparent: " << info.parent_name
-       << " (pool " << info.parent_pool << ")"
-       << std::endl;
+  string s = ""; 
+
+  if (features & RBD_FEATURE_LAYERING)
+    s += "layering";
+  return s;
 }
 
 struct MyProgressContext : public librbd::ProgressContext {
@@ -166,14 +167,36 @@ static int do_list(librbd::RBD &rbd, librados::IoCtx& io_ctx)
 }
 
 static int do_create(librbd::RBD &rbd, librados::IoCtx& io_ctx,
-		     const char *imgname, uint64_t size, int *order)
+		     const char *imgname, uint64_t size, int *order,
+		     bool old_format, uint64_t features)
 {
-  int r = rbd.create(io_ctx, imgname, size, order);
+  int r;
+  if (features == 0) 
+    features = RBD_FEATURES_ALL;
+
+  if (old_format)
+    r = rbd.create(io_ctx, imgname, size, order);
+  else
+    r = rbd.create2(io_ctx, imgname, size, features, order);
   if (r < 0)
     return r;
   return 0;
 }
 
+static int do_clone(librbd::RBD &rbd, librados::IoCtx &p_ioctx,
+		    const char *p_name, const char *p_snapname, 
+		    librados::IoCtx &c_ioctx, const char *c_name,
+		    uint64_t features, int *c_order)
+{
+  if (features == 0)
+    features = RBD_FEATURES_ALL;
+  else if ((features & RBD_FEATURE_LAYERING) != RBD_FEATURE_LAYERING)
+    return -EINVAL;
+
+  return rbd.clone(p_ioctx, p_name, p_snapname, c_ioctx, c_name, features,
+		    c_order);
+}
+
 static int do_rename(librbd::RBD &rbd, librados::IoCtx& io_ctx,
 		     const char *imgname, const char *destname)
 {
@@ -186,11 +209,47 @@ static int do_rename(librbd::RBD &rbd, librados::IoCtx& io_ctx,
 static int do_show_info(const char *imgname, librbd::Image& image)
 {
   librbd::image_info_t info;
+  string parent_pool, parent_name, parent_snapname;
+  uint8_t old_format;
+  uint64_t overlap, features;
   int r = image.stat(info, sizeof(info));
   if (r < 0)
     return r;
 
-  print_info(imgname, info);
+  r = image.old_format(&old_format);
+  if (r < 0)
+    return r;
+
+  r = image.overlap(&overlap);
+  if (r < 0)
+    return r;
+
+  r = image.features(&features);
+  if (r < 0)
+    return r;
+
+  cout << "rbd image '" << imgname << "':\n"
+       << "\tsize " << prettybyte_t(info.size) << " in "
+       << info.num_objs << " objects"
+       << std::endl
+       << "\torder " << info.order
+       << " (" << prettybyte_t(info.obj_size) << " objects)"
+       << std::endl
+       << "\tblock_name_prefix: " << info.block_name_prefix
+       << std::endl
+       << "\told format: " << (old_format ? "True" : "False")
+       << std::endl
+       << "\tfeatures: " << feature_str(features)
+       << std::endl;
+
+  // parent info, if present 
+  if ((image.parent_info(&parent_pool, &parent_name, &parent_snapname) == 0) &&
+      parent_name.length() > 0) {
+
+    cout << "\tparent: " << parent_pool << "/" << parent_name
+	 << "@" << parent_snapname << std::endl;
+    cout << "\toverlap: " << prettybyte_t(overlap) << std::endl;
+  }
   return 0;
 }
 
@@ -225,6 +284,7 @@ static int do_list_snaps(librbd::Image& image)
   if (r < 0)
     return r;
 
+  cout << "ID\tNAME\t\tSIZE" << std::endl;
   for (std::vector<librbd::snap_info_t>::iterator i = snaps.begin(); i != snaps.end(); ++i) {
     cout << i->id << '\t' << i->name << '\t' << i->size << std::endl;
   }
@@ -402,16 +462,20 @@ done_img:
 }
 
 static int do_import(librbd::RBD &rbd, librados::IoCtx& io_ctx,
-		     const char *imgname, int *order, const char *path)
+		     const char *imgname, int *order, const char *path,
+		     bool old_format, uint64_t features, int64_t size)
 {
-  int fd = open(path, O_RDONLY);
-  int r;
-  int64_t size = 0;
+  int fd, r;
   struct stat stat_buf;
-  string md_oid;
   struct fiemap *fiemap;
   MyProgressContext pc("Importing image");
 
+  if (! strcmp(path, "-")) {
+    fd = 0;
+  } else {
+    fd = open(path, O_RDONLY);
+  }
+
   if (fd < 0) {
     r = -errno;
     cerr << "error opening " << path << std::endl;
@@ -424,7 +488,9 @@ static int do_import(librbd::RBD &rbd, librados::IoCtx& io_ctx,
     cerr << "stat error " << path << std::endl;
     return r;
   }
-  size = (uint64_t)stat_buf.st_size;
+  if (stat_buf.st_size)
+    size = (uint64_t)stat_buf.st_size;
+
   if (!size) {
     r = get_block_device_size(fd, &size);
     if (r < 0) {
@@ -435,10 +501,7 @@ static int do_import(librbd::RBD &rbd, librados::IoCtx& io_ctx,
 
   assert(imgname);
 
-  md_oid = imgname;
-  md_oid += RBD_SUFFIX;
-
-  r = do_create(rbd, io_ctx, imgname, size, order);
+  r = do_create(rbd, io_ctx, imgname, size, order, old_format, features);
   if (r < 0) {
     cerr << "image creation failed" << std::endl;
     return r;
@@ -505,7 +568,12 @@ static int do_import(librbd::RBD &rbd, librados::IoCtx& io_ctx,
       while (cur_seg) {
         bufferptr p(cur_seg);
         //cerr << "reading " << cur_seg << " bytes at offset " << file_pos << std::endl;
-        ssize_t rval = TEMP_FAILURE_RETRY(::pread(fd, p.c_str(), cur_seg, file_pos));
+	ssize_t rval;
+	if(extent == 0 && fiemap->fm_extents[extent].fe_logical == 0) {
+	  rval = TEMP_FAILURE_RETRY(::read(fd, p.c_str(), cur_seg));
+	} else {
+	  rval = TEMP_FAILURE_RETRY(::pread(fd, p.c_str(), cur_seg, file_pos));
+	}
         if (rval < 0) {
           r = -errno;
           cerr << "error reading file: " << cpp_strerror(r) << std::endl;
@@ -580,10 +648,22 @@ static int do_watch(librados::IoCtx& pp, const char *imgname)
   uint64_t cookie;
   RbdWatchCtx ctx(imgname);
 
-  md_oid = imgname;
-  md_oid += RBD_SUFFIX;
+  string old_header_oid = imgname;
+  old_header_oid += RBD_SUFFIX;
+  string new_header_oid = RBD_HEADER_PREFIX;
+  new_header_oid += imgname;
+  bool old_format = true;
 
-  int r = pp.watch(md_oid, 0, &cookie, &ctx);
+  int r = pp.stat(old_header_oid, NULL, NULL);
+  if (r < 0) {
+    r = pp.stat(new_header_oid, NULL, NULL);
+    if (r < 0)
+      return r;
+    old_format = false;
+  }
+
+  r = pp.watch(old_format ? old_header_oid : new_header_oid,
+	       0, &cookie, &ctx);
   if (r < 0) {
     cerr << "watch failed" << std::endl;
     return r;
@@ -600,7 +680,7 @@ static int do_kernel_add(const char *poolname, const char *imgname,
 			 const char *user)
 {
   MonMap monmap;
-  int r = MonClient::build_initial_monmap(g_ceph_context, monmap);
+  int r = monmap.build_initial(g_ceph_context, cerr);
   if (r < 0)
     return r;
 
@@ -614,7 +694,7 @@ static int do_kernel_add(const char *poolname, const char *imgname,
 
   oss << " name=" << user;
 
-  char key_name[strlen(user) + strlen("client.")];
+  char key_name[strlen(user) + strlen("client.")+1];
   snprintf(key_name, sizeof(key_name), "client.%s", user);
   char secret_buf[MAX_SECRET_LEN];
   char *secret = NULL;
@@ -833,6 +913,7 @@ enum {
   OPT_LIST,
   OPT_INFO,
   OPT_CREATE,
+  OPT_CLONE,
   OPT_RESIZE,
   OPT_RM,
   OPT_EXPORT,
@@ -860,6 +941,8 @@ static int get_cmd(const char *cmd, bool snapcmd)
       return OPT_INFO;
     if (strcmp(cmd, "create") == 0)
       return OPT_CREATE;
+    if (strcmp(cmd, "clone") == 0)
+      return OPT_CLONE;
     if (strcmp(cmd, "resize") == 0)
       return OPT_RESIZE;
     if (strcmp(cmd, "rm") == 0)
@@ -931,7 +1014,9 @@ int main(int argc, const char **argv)
   const char *poolname = NULL;
   uint64_t size = 0;  // in bytes
   int order = 0;
-  const char *imgname = NULL, *snapname = NULL, *destname = NULL, *dest_poolname = NULL, *path = NULL, *secretfile = NULL, *user = NULL, *devpath = NULL;
+  bool old_format = true;
+  uint64_t features = RBD_FEATURE_LAYERING;
+  const char *imgname = NULL, *snapname = NULL, *destname = NULL, *dest_poolname = NULL, *dest_snapname = NULL, *path = NULL, *secretfile = NULL, *user = NULL, *devpath = NULL;
 
   std::string val;
   std::ostringstream err;
@@ -943,6 +1028,8 @@ int main(int argc, const char **argv)
     } else if (ceph_argparse_flag(args, i, "-h", "--help", (char*)NULL)) {
       usage();
       exit(0);
+    } else if (ceph_argparse_flag(args, i, "--new-format", (char*)NULL)) {
+      old_format = false;
     } else if (ceph_argparse_witharg(args, i, &val, "-p", "--pool", (char*)NULL)) {
       poolname = strdup(val.c_str());
     } else if (ceph_argparse_witharg(args, i, &val, "--dest-pool", (char*)NULL)) {
@@ -970,6 +1057,8 @@ int main(int argc, const char **argv)
       secretfile = strdup(val.c_str());
     } else if (ceph_argparse_witharg(args, i, &val, "--user", (char*)NULL)) {
       user = strdup(val.c_str());
+    } else if (ceph_argparse_witharg(args, i, &val, "--parent", (char *)NULL)) {
+      imgname = strdup(val.c_str());
     } else {
       ++i;
     }
@@ -1028,6 +1117,13 @@ int main(int argc, const char **argv)
       case OPT_RENAME:
 	set_conf_param(v, &imgname, &destname);
 	break;
+      case OPT_CLONE:
+	if (imgname == NULL) {
+	  set_conf_param(v, &imgname, NULL);
+        } else {
+	  set_conf_param(v, &destname, NULL);
+	}
+	break;
       case OPT_SHOWMAPPED:
 	usage_exit();
 	break;
@@ -1065,17 +1161,17 @@ int main(int argc, const char **argv)
   if (snapname && opt_cmd != OPT_SNAP_CREATE && opt_cmd != OPT_SNAP_ROLLBACK &&
       opt_cmd != OPT_SNAP_REMOVE && opt_cmd != OPT_INFO &&
       opt_cmd != OPT_EXPORT && opt_cmd != OPT_COPY &&
-      opt_cmd != OPT_MAP) {
+      opt_cmd != OPT_MAP && opt_cmd != OPT_CLONE) {
     cerr << "error: snapname specified for a command that doesn't use it" << std::endl;
     usage_exit();
   }
   if ((opt_cmd == OPT_SNAP_CREATE || opt_cmd == OPT_SNAP_ROLLBACK ||
-       opt_cmd == OPT_SNAP_REMOVE) && !snapname) {
+       opt_cmd == OPT_SNAP_REMOVE || opt_cmd == OPT_CLONE) && !snapname) {
     cerr << "error: snap name was not specified" << std::endl;
     usage_exit();
   }
 
-  set_pool_image_name(dest_poolname, destname, (char **)&dest_poolname, (char **)&destname, NULL);
+  set_pool_image_name(dest_poolname, destname, (char **)&dest_poolname, (char **)&destname, (char **)&dest_snapname);
 
   if (!poolname)
     poolname = "rbd";
@@ -1085,11 +1181,21 @@ int main(int argc, const char **argv)
   if (opt_cmd == OPT_EXPORT && !path)
     path = imgname;
 
-  if (opt_cmd == OPT_COPY && !destname ) {
+  if ((opt_cmd == OPT_COPY || opt_cmd == OPT_CLONE) && !destname ) {
     cerr << "error: destination image name was not specified" << std::endl;
     usage_exit();
   }
 
+  if ((opt_cmd == OPT_CLONE) && dest_snapname) {
+    cerr << "error: cannot clone to a snapshot" << std::endl;
+    usage_exit();
+  }
+
+  if ((opt_cmd == OPT_CLONE) && size) {
+    cerr << "error: clone must begin at size of parent" << std::endl;
+    usage_exit();
+  }
+
   if ((opt_cmd == OPT_RENAME) && (strcmp(poolname, dest_poolname) != 0)) {
     cerr << "error: mv/rename across pools not supported" << std::endl;
     cerr << "source pool: " << poolname << " dest pool: " << dest_poolname
@@ -1140,7 +1246,7 @@ int main(int argc, const char **argv)
     }
   }
 
-  if (opt_cmd == OPT_COPY || opt_cmd == OPT_IMPORT) {
+  if (opt_cmd == OPT_COPY || opt_cmd == OPT_IMPORT || opt_cmd == OPT_CLONE) {
     r = rados.ioctx_create(dest_poolname, dest_io_ctx);
     if (r < 0) {
       cerr << "error opening pool " << dest_poolname << ": " << cpp_strerror(-r) << std::endl;
@@ -1174,13 +1280,28 @@ int main(int argc, const char **argv)
       usage();
       exit(1);
     }
-    r = do_create(rbd, io_ctx, imgname, size, &order);
+    r = do_create(rbd, io_ctx, imgname, size, &order, old_format, features);
     if (r < 0) {
       cerr << "create error: " << cpp_strerror(-r) << std::endl;
       exit(1);
     }
     break;
 
+  case OPT_CLONE:
+    if (order && (order < 12 || order > 25)) {
+      cerr << "order must be between 12 (4 KB) and 25 (32 MB)" << std::endl;
+      usage();
+      exit(1);
+    }
+
+    r = do_clone(rbd, io_ctx, imgname, snapname, dest_io_ctx, destname,
+		 features, &order);
+    if (r < 0) {
+      cerr << "clone error: " << cpp_strerror(-r) << std::endl;
+      exit(1);
+    }
+    break;
+
   case OPT_RENAME:
     r = do_rename(rbd, io_ctx, imgname, destname);
     if (r < 0) {
@@ -1299,11 +1420,12 @@ int main(int argc, const char **argv)
     break;
 
   case OPT_IMPORT:
-     if (!path) {
+    if (!path) {
       cerr << "pathname should be specified" << std::endl;
       exit(1);
     }
-    r = do_import(rbd, dest_io_ctx, destname, &order, path);
+    r = do_import(rbd, dest_io_ctx, destname, &order, path,
+		  old_format, features, size);
     if (r < 0) {
       cerr << "import failed: " << cpp_strerror(-r) << std::endl;
       exit(1);
