@@ -22,7 +22,7 @@ MON_ADDR=""
 
 conf="ceph.conf"
 
-keyring_fn="keyring"
+keyring_fn="$PWD/keyring"
 osdmap_fn="/tmp/ceph_osdmap.$$"
 monmap_fn="/tmp/ceph_monmap.$$"
 
@@ -165,7 +165,8 @@ else
         debug osd = 25
         debug monc = 20
         debug journal = 20
-        debug filestore = 20'
+        debug filestore = 20
+        debug objclass = 20'
     CMDSDEBUG='
         lockdep = 1
         debug ms = 1
@@ -200,14 +201,14 @@ test -d out || mkdir out
 $SUDO rm -rf out/*
 test -d gmon && $SUDO rm -rf gmon/*
 
-[ "$cephx" -eq 1 ] && test -e $keyring_fn && rm $keyring_fn
+[ "$cephx" -eq 1 ] && [ "$new" -eq 1 ] && test -e $keyring_fn && rm $keyring_fn
 
 
 # figure machine's ip
+HOSTNAME=`hostname`
 if [ "$localhost" -eq 1 ]; then
     IP="127.0.0.1"
 else
-    HOSTNAME=`hostname`
     echo hostname $HOSTNAME
     RAW_IP=`hostname --ip-address`
     # filter out IPv6 and localhost addresses
@@ -260,7 +261,6 @@ if [ "$start_mon" -eq 1 ]; then
 		        cat <<EOF > $conf
 ; generated by vstart.sh on `date`
 [global]
-        keyring = keyring
         osd pg bits = 3
         osd pgp bits = 5  ; (invalid, but ceph should cope!)
 $extra_conf
@@ -269,14 +269,22 @@ EOF
         auth supported = cephx
 EOF
 			cat <<EOF >> $conf
+
+[client]
+        keyring = $keyring_fn
+
 [mds]
 $DAEMONOPTS
 $CMDSDEBUG
         mds debug frag = true
         mds debug auth pins = true
         mds debug subtrees = true
+        mds data = dev/mds.\$id
 [osd]
 $DAEMONOPTS
+        osd data = dev/osd\$id
+        osd journal = dev/osd\$id.journal
+        osd journal size = 100
         osd class tmp = out
         osd class dir = .libs
         osd scrub load threshold = 5.0
@@ -285,7 +293,7 @@ $COSDDEBUG
 [mon]
 $DAEMONOPTS
 $CMONDEBUG
-
+        mon cluster log file = out/cluster.mon.\$id.log
 EOF
 		fi
 
@@ -298,8 +306,8 @@ EOF
 			echo
 		fi
 
-	        [ "$cephx" -eq 1 ] && $SUDO $CEPH_BIN/ceph-authtool --create-keyring --gen-key --name=mon. $keyring_fn
-	        [ "$cephx" -eq 1 ] && $SUDO $CEPH_BIN/ceph-authtool --gen-key --name=client.admin --set-uid=0 \
+	        $SUDO $CEPH_BIN/ceph-authtool --create-keyring --gen-key --name=mon. $keyring_fn
+	        $SUDO $CEPH_BIN/ceph-authtool --gen-key --name=client.admin --set-uid=0 \
 		    --cap mon 'allow *' \
 		    --cap osd 'allow *' \
 		    --cap mds allow \
@@ -314,6 +322,7 @@ EOF
 			if [ $overwrite_conf -eq 1 ]; then
 				cat <<EOF >> $conf
 [mon.$f]
+        host = $HOSTNAME
         mon data = dev/mon.$f
         mon addr = $IP:$(($CEPH_PORT+$count))
 EOF
@@ -326,6 +335,9 @@ EOF
 
 		for f in $MONS
 		do
+		    cmd="rm -rf dev/mon.$f"
+		    echo $cmd
+		    $cmd
 		    cmd="$CEPH_BIN/ceph-mon --mkfs -c $conf -i $f --monmap=$monmap_fn"
 		    [ "$cephx" -eq 1 ] && cmd="$cmd --keyring=$keyring_fn"
 		    echo $cmd
@@ -352,25 +364,22 @@ if [ "$start_osd" -eq 1 ]; then
 	    if [ $overwrite_conf -eq 1 ]; then
 		    cat <<EOF >> $conf
 [osd.$osd]
-        osd data = dev/osd$osd
-        osd journal = dev/osd$osd.journal
-        osd journal size = 100
-EOF
-		    [ "$cephx" -eq 1 ] && cat <<EOF >> $conf
-        keyring = dev/osd$osd/keyring
+        host = $HOSTNAME
 EOF
+		    rm -rf dev/osd$osd || true
+		    for f in dev/osd$osd/* ; do btrfs sub delete $f || true ; done || true
+		    mkdir -p dev/osd$osd
 	    fi
 
-	    echo add osd$osd
-	    $SUDO $CEPH_ADM osd create $osd
-	    $SUDO $CEPH_ADM osd crush add $osd osd.$osd 1.0 host=localhost rack=localrack pool=default
-	    $SUDO $CEPH_BIN/ceph-osd -i $osd $ARGS --mkfs --mkkey
+	    uuid=`uuidgen`
+	    echo "add osd$osd $uuid"
+	    $SUDO $CEPH_ADM osd create $uuid
+	    $SUDO $CEPH_ADM osd crush set $osd osd.$osd 1.0 host=localhost rack=localrack pool=default
+	    $SUDO $CEPH_BIN/ceph-osd -i $osd $ARGS --mkfs --mkkey --osd-uuid $uuid
 
-	    if [ "$cephx" -eq 1 ]; then
-		key_fn=dev/osd$osd/keyring
-		echo adding osd$osd key to auth repository
-		$SUDO $CEPH_ADM -i $key_fn auth add osd.$osd osd "allow *" mon "allow rwx"
-	    fi
+	    key_fn=dev/osd$osd/keyring
+	    echo adding osd$osd key to auth repository
+	    $SUDO $CEPH_ADM -i $key_fn auth add osd.$osd osd "allow *" mon "allow rwx"
 	fi
 	echo start osd$osd
 	run 'osd' $SUDO $CEPH_BIN/ceph-osd -i $osd $ARGS $COSD_ARGS
@@ -391,21 +400,18 @@ if [ "$start_mds" -eq 1 ]; then
     for name in a b c d e f g h i j k l m n o p
     do
 	if [ "$new" -eq 1 ]; then
-	    key_fn=dev/mds.$name.keyring
+	    mkdir -p dev/mds.$name
+	    key_fn=dev/mds.$name/keyring
 	    if [ $overwrite_conf -eq 1 ]; then
 	    	cat <<EOF >> $conf
 [mds.$name]
+        host = $HOSTNAME
 EOF
-		if [ "$cephx" -eq 1 ]; then
-	    	    cat <<EOF >> $conf
-        keyring = $key_fn
-EOF
-		fi
 		if [ "$standby" -eq 1 ]; then
+		    mkdir -p dev/mds.${name}s
 		    cat <<EOF >> $conf
        mds standby for rank = $mds
 [mds.${name}s]
-        keyring = dev/mds.${name}s.keyring
         mds standby replay = true
         mds standby for name = ${name}
 EOF
@@ -415,8 +421,8 @@ EOF
 	    $SUDO $CEPH_ADM -i $key_fn auth add mds.$name mon 'allow *' osd 'allow *' mds 'allow'
 	    if [ "$standby" -eq 1 ]; then
 		    $SUDO $CEPH_BIN/ceph-authtool --create-keyring --gen-key --name=mds.${name}s \
-			dev/mds.${name}s.keyring
-                    $SUDO $CEPH_ADM -i dev/mds.${name}s.keyring auth add mds.${name}s \
+			dev/mds.${name}s/keyring
+                    $SUDO $CEPH_ADM -i dev/mds.${name}s/keyring auth add mds.${name}s \
 			mon 'allow *' osd 'allow *' mds 'allow'
 	    fi
 	fi
